from __future__ import annotations

import itertools

from prettyqt import core, custom_widgets, widgets
from prettyqt.utils import listdelegators


class MultiLineLayout(widgets.BoxLayout):
    """Nested Boxlayout."""

    def __init__(self, vertical: bool = True, row_number: int = 3, **kwargs):
        self.row_nb = row_number
        self.layouts = []
        self._sub_layout_type = "box"
        direction = self.Direction.TopToBottom if vertical else self.Direction.LeftToRight
        super().__init__(direction, **kwargs)

    def set_direction(
        self, direction: widgets.BoxLayout.Direction | widgets.boxlayout.DirectionStr
    ):
        super().set_direction(direction)
        direction = self.get_sub_direction()
        for layout in self.layouts:
            if isinstance(layout, widgets.QBoxLayout | MultiLineLayout):
                layout.set_direction(direction)

    def get_sub_direction(self) -> widgets.BoxLayout.Direction:
        return (
            self.Direction.LeftToRight
            if self.get_direction() == "top_to_bottom"
            else self.Direction.TopToBottom
        )

    def _add_sub_layout(self):
        match self._sub_layout_type:
            case "box":
                direction = self.get_sub_direction()
                layout = widgets.BoxLayout(direction)
            case "grid":
                layout = widget.GridLayout()
            case "flow":
                layout = custom_widgets.FlowLayout()
            case "stacked":
                layout = widgets.StackedLayout()
            case "nested":
                direction = self.get_sub_direction()
                layout = MultiLineLayout(direction)
        super().addLayout(layout)
        self.layouts.append(layout)

    def addWidget(self, widget: widgets.QWidget, **kwargs):
        if not self.layouts or self.layouts[-1].count() == self.row_nb:
            self._add_sub_layout()
        self.layouts[-1].add(widget, **kwargs)

    def addLayout(self, layout: widgets.QLayout):
        if not self.layouts or self.layouts[-1].count() == self.row_nb:
            self._add_sub_layout()
        self.layouts[-1].add(layout)

    def addItem(self, item):
        if not self.layouts or self.layouts[-1].count() == self.row_nb:
            self._add_sub_layout()
        self.layouts[-1].add(item)

    def get_items(self) -> listdelegators.ListDelegator[widgets.QLayoutItem]:
        items = [i.get_items() for i in self.layouts]
        return listdelegators.ListDelegator(itertools.chain(*items))

    def itemAt(self, idx: int) -> widgets.QLayoutItem | None:
        if len(self.layouts) == 0 or len(self.get_items()) == 0:
            raise IndexError(idx)
        # doesnt seem right?
        return None if idx == len(self.get_items()) else self.get_items()[idx]

    def takeAt(self, idx: int) -> widgets.QLayoutItem | None:  # or 0 according to docs?
        layout_idx, item_idx = divmod(idx, len(self.layouts))
        layout = self.layouts[layout_idx]
        item = layout.takeAt(item_idx)
        for i, layout in enumerate(self.layouts[layout_idx:-1], start=layout_idx):
            item = self.layouts[i + 1].takeAt(0)
            layout.addItem(item)
        if len(self.layouts[-1]) == 0:
            super().takeAt(super().count() - 1)
        return item

    def count(self) -> int:
        return len(self.get_items())

    def indexOf(self, item: widgets.QLayoutItem) -> int:
        return self.get_items().index(item)

    def set_sub_layout_type(self, layout_type: str):
        self._sub_layout_type = layout_type

    def get_sub_layout_type(self) -> str:
        return self._sub_layout_type

    sub_layout_type = core.Property(
        str,
        get_sub_layout_type,
        set_sub_layout_type,
        doc="Layout for the sub layouts",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.Widget()
    layout = MultiLineLayout(parent=widget)
    layout.addWidget(widgets.PushButton("Short"))
    layout.addWidget(widgets.PushButton("Longer"))
    layout.addWidget(widgets.PushButton("Different text"))
    layout.addWidget(widgets.PushButton("More text"))
    layout.addWidget(widgets.PushButton("Even longer button text"))
    layout.addWidget(widgets.PushButton("Even longer button text"))
    layout.addWidget(widgets.PushButton("Last one"))
    layout.takeAt(0)
    # layout.set_direction("left_to_right")
    # print(layout.items())
    # print(layout.itemAt(5))
    widget.set_layout(layout)
    widget.show()
    app.exec()



"""Tests for `prettyqt` package."""

import pytest


scxml = pytest.importorskip("prettyqt.scxml")


def test_scxmlcompiler(qtbot):
    compiler = scxml.ScxmlCompiler()
    compiler.get_file_path()


def test_scxmlstatemachine(qtbot):
    machine = scxml.ScxmlStateMachine()
    assert machine



from __future__ import annotations

import logging

from prettyqt import core
from prettyqt.utils import helpers


logger = logging.getLogger(__name__)


class SliceIdentityProxyModel(core.IdentityProxyModel):
    """Base model for slice proxy models.

    By itself, this proxy does nothing.
    It just provides all needed helper functions to deal with slices.
    `indexer_contains` should be just by subclasses in order to check whether the proxy
    should be applied.
    """

    def __init__(self, indexer=None, **kwargs):
        super().__init__(**kwargs)
        self._indexer = (slice(None), slice(None))
        self.set_indexer(indexer)

    def indexer_contains(self, index: core.ModelIndex | tuple[int, int]) -> bool:
        """Check whether given ModelIndex is included in our Indexer."""
        if isinstance(index, core.ModelIndex):
            index = (index.row(), index.column())
        col_slice = self.get_column_slice()
        row_slice = self.get_row_slice()
        source = self.sourceModel()
        col_slice = self.update_slice_boundaries(col_slice, count=source.columnCount())
        row_slice = self.update_slice_boundaries(row_slice, count=source.rowCount())
        # logger.info(f"{col_slice=} {row_slice=}")
        to_check = (row_slice, col_slice)  # instead of _indexer, for negative indexes.
        return helpers.is_position_in_index(*index, to_check)

    def update_slice_boundaries(self, sl: slice, count: int) -> slice:
        """Update slice boundaries by resolving negative indexes."""
        # Not sure yet whats the best approach here and which cases I should support...
        # if sl.end is larger than count, clip it (or perhaps throw exception?)
        # if sl.stop is not None and sl.stop >= count:
        #     sl = slice(sl.start, count, sl.step)
        # resolve negative start value
        if sl.start is not None and sl.start < 0:
            start = count + sl.start
            end = count + sl.stop
            # end = start + (sl.stop - sl.start)
            if start < 0:
                raise IndexError(sl.start)
            sl = slice(start, end, sl.step)
        # if sl.stop is not None and sl.stop < 0:
        #     stop = source.columnCount() + sl.stop
        #     if stop < 0:
        #         raise IndexError(sl.stop)
        #     sl = slice(sl.start, stop, sl.step)
        return sl

    def set_indexer(self, indexer):
        """Takes basically anything which is common to use for __getitem__."""
        match indexer:
            case None:
                self._indexer = (slice(None), slice(None))
            case int() as column:
                self.set_column_slice(column)
            case slice() as col_slice:
                self._indexer = (col_slice, slice(None))
            case row_slice, col_slice:
                self.set_column_slice(col_slice)
                self.set_row_slice(row_slice)
            case _:
                raise TypeError(indexer)

    def get_column_slice(self) -> slice:
        match self._indexer:
            case None | (_, None):
                return slice(None)
            case (_, slice() as col_slice):
                return col_slice
            case _:
                raise TypeError(self._indexer)

    def get_row_slice(self) -> slice:
        match self._indexer:
            case None | (None, _):
                return slice(None)
            case (slice() as row_slice, _):
                return row_slice
            case _:
                raise TypeError(self._indexer)

    def set_column_slice(
        self, value: slice | int | None | tuple[int | None, int | None, int | None]
    ):
        """Throw anything at this method in order to set the column slice."""
        match value:
            case slice() as col_slice:
                sl = col_slice
            case int() as col:
                sl = slice(col, col + 1)
            case None:
                sl = slice(None)
            case (
                int() | None as start,
                int() | None as stop,
                int() | None as step,
            ):
                sl = slice(start, stop, step)
            case _:
                raise TypeError(value)
        self._indexer = (self.get_row_slice(), sl)

    def set_row_slice(
        self, value: slice | int | None | tuple[int | None, int | None, int | None]
    ):
        """Throw anything at this method in order to set the row slice."""
        match value:
            case slice() as row_slice:
                sl = row_slice
            case int() as row:
                sl = slice(row, row + 1)
            case None:
                sl = slice(None)
            case (
                int() | None as start,
                int() | None as stop,
                int() | None as step,
            ):
                sl = slice(start, stop, step)
            case _:
                raise TypeError(value)
        self._indexer = (sl, self.get_column_slice())

    def get_row_range(self) -> range:
        """Return a range for the row slice with valid start / stop / step values."""
        sl = self.get_row_slice()
        row_count = self.sourceModel().rowCount()
        return range(sl.start or 0, sl.stop or row_count, sl.step or 1)

    def get_column_range(self) -> range:
        """Return a range for the column slice with valid start / stop / step values."""
        sl = self.get_column_slice()
        col_count = self.sourceModel().columnCount()
        return range(sl.start or 0, sl.stop or col_count, sl.step or 1)

    def position_in_column_slice(self, col: int) -> int:
        """Can be interpreted as slice.index(col) if slice would be a list."""
        sl = self.get_column_slice()
        return int((col - (sl.start or 0)) / (sl.step or 1))

    def position_in_row_slice(self, row: int) -> int:
        """Can be interpreted as slice.index(row) if slice would be a list."""
        sl = self.get_row_slice()
        return int((row - (sl.start or 0)) / (sl.step or 1))

    # The Qt typesystems dont like slices (or ranges / tuples)
    # seems what works is to throw tuples at a QtProperty declared as list.
    # otherwise change this to list getters/setters.

    def get_column_tuple(self) -> tuple[int | None, int | None, int | None]:
        """Get tuple representation of the column slice."""
        sl = self.get_column_slice()
        return (sl.start, sl.stop, sl.step)

    def get_row_tuple(self) -> tuple[int | None, int | None, int | None]:
        """Get tuple representation of the row slice."""
        sl = self.get_row_slice()
        return (sl.start, sl.stop, sl.step)

    column_slice = core.Property(
        list,
        get_column_tuple,
        set_column_slice,
        doc="Column slice to include for the proxy",
    )
    row_slice = core.Property(
        list,
        get_row_tuple,
        set_column_slice,
        doc="Row slice to include for the proxy",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    data = dict(
        first=["John", "Mary"],
        last=["Doe", "Bo"],
        height=[5.5, 6.0],
        weight=[130, 150],
    )
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[-1].change_flags(enabled=False)
    model.insertColumn(0, [gui.StandardItem("new"), gui.StandardItem("new")])
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import os
import pathlib
from typing import Any

from prettyqt import core
from prettyqt.utils import get_repr


class FileInfo(core.QFileInfo):
    """System-independent file information."""

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], os.PathLike):
            super().__init__(os.fspath(args[0]))
        else:
            super().__init__(*args, **kwargs)

    def __repr__(self):
        return get_repr(self, self.absoluteFilePath())

    @property
    def _absoluteFilePath(self) -> str:
        return self.absoluteFilePath()

    __match_args__ = ("_absoluteFilePath",)

    def __str__(self):
        return self.absoluteFilePath()

    def __fspath__(self) -> str:
        return self.absoluteFilePath()

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.get_absolute_file_path(), attr)

    def __reduce__(self):
        return type(self), (self.absoluteFilePath(),)

    def get_dir(self) -> pathlib.Path:
        return pathlib.Path(self.dir().absolutePath())

    def get_absolute_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.absoluteFilePath())

    def get_birth_time(self) -> core.DateTime:
        return core.DateTime(self.birthTime())

    def get_metadata_change_time(self) -> core.DateTime:
        return core.DateTime(self.metadataChangeTime())

    def get_last_modified(self) -> core.DateTime:
        return core.DateTime(self.lastModified())

    def get_last_read(self) -> core.DateTime:
        return core.DateTime(self.lastRead())


if __name__ == "__main__":
    p = pathlib.Path.home()
    f = FileInfo(p)



from __future__ import annotations

from typing import Self

from prettyqt import core


class Time(core.QTime):
    """Clock time functions."""

    def __repr__(self):
        super_repr = super().__repr__()
        template = str(super_repr).split("(")[1]  # type: ignore
        return f"{type(self).__name__}({template}"

    def __str__(self):
        return self.toString()

    def __reduce__(self):
        return type(self), (self.hour(), self.minute(), self.second(), self.msec())

    def __format__(self, format_spec: str):
        return self.toString(format_spec)

    @property
    def _hour(self) -> int:
        return self.hour()

    @property
    def _minute(self) -> int:
        return self.minute()

    @property
    def _second(self) -> int:
        return self.second()

    @property
    def _msec(self) -> int:
        return self.msec()

    __match_args__ = ("_hour", "_minute", "_second", "_msec")

    def add_msecs(self, msecs: int) -> Self:
        return type(self)(self.addMSecs(msecs))

    def add_secs(self, secs: int) -> Self:
        return type(self)(self.addSecs(secs))

    @classmethod
    def get_current_time(cls) -> Self:
        return cls(cls.currentTime())

    @classmethod
    def from_string(cls, *args, **kwargs) -> Self:
        return cls(cls.fromString(*args, **kwargs))


if __name__ == "__main__":
    import datetime

    time = Time(datetime.time(10, 10, 10))
    print(time)



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict


SliderActionStr = Literal[
    "none",
    "single_step_add",
    "single_step_sub",
    "page_step_add",
    "page_step_sub",
    "to_minimum",
    "to_maximum",
    "move",
]

SLIDER_ACTION: bidict[SliderActionStr, widgets.QAbstractSlider.SliderAction] = bidict(
    none=widgets.QAbstractSlider.SliderAction.SliderNoAction,
    single_step_add=widgets.QAbstractSlider.SliderAction.SliderSingleStepAdd,
    single_step_sub=widgets.QAbstractSlider.SliderAction.SliderSingleStepSub,
    page_step_add=widgets.QAbstractSlider.SliderAction.SliderPageStepAdd,
    page_step_sub=widgets.QAbstractSlider.SliderAction.SliderPageStepSub,
    to_minimum=widgets.QAbstractSlider.SliderAction.SliderToMinimum,
    to_maximum=widgets.QAbstractSlider.SliderAction.SliderToMaximum,
    move=widgets.QAbstractSlider.SliderAction.SliderMove,
)


class AbstractSliderMixin(widgets.WidgetMixin):
    value_changed = core.Signal(int)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "orientation": constants.ORIENTATION,
        }
        return maps

    def on_value_change(self):
        self.value_changed.emit(self.value())

    def is_horizontal(self) -> bool:
        """Check if silder is horizontal.

        Returns:
            True if horizontal, else False
        """
        return self.orientation() == constants.HORIZONTAL

    def is_vertical(self) -> bool:
        """Check if silder is vertical.

        Returns:
            True if vertical, else False
        """
        return self.orientation() == constants.VERTICAL

    def set_horizontal(self):
        """Set slider orientation to horizontal."""
        self.setOrientation(constants.HORIZONTAL)

    def set_vertical(self):
        """Set slider orientation to vertical."""
        self.setOrientation(constants.VERTICAL)

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the slider.

        Args:
            orientation: orientation for the slider
        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def scroll_to_min(self):
        """Scroll to the minimum value of the slider."""
        self.setValue(self.minimum())

    def scroll_to_max(self):
        """Scroll to the maximum value of the slider."""
        self.setValue(self.maximum())

    def set_range(self, min_val: int, max_val: int):
        self.setRange(min_val, max_val)

    def set_step_size(self, step_size: int):
        self.setSingleStep(step_size)

    def set_repeat_action(
        self,
        action: SliderActionStr | widgets.QAbstractSlider.SliderAction,
        threshold: int = 500,
        repeat_time: int = 50,
    ):
        """Set the repeat action.

        Args:
            action: repeat action
            threshold: initial delay in ms
            repeat_time: repeat time in ms
        """
        self.setRepeatAction(SLIDER_ACTION.get_enum_value(action), threshold, repeat_time)

    def get_repeat_action(self) -> SliderActionStr:
        """Get current repeat action.

        Returns:
            current repeat action
        """
        return SLIDER_ACTION.inverse[self.repeatAction()]

    def trigger_action(
        self, action: SliderActionStr | widgets.QAbstractSlider.SliderAction
    ):
        """Trigger slider action."""
        self.triggerAction(SLIDER_ACTION.get_enum_value(action))

    def get_value(self) -> int:
        return self.value()

    def set_value(self, value: int):
        self.setValue(value)

    def on_scrollbar_range_changed(self, minval: int, maxval: int):
        if self.value() >= self.maximum() - 1:
            self.setValue(maxval)

    def set_auto_scroll_to_end(self, scroll: bool = True):
        """Set to always scroll to the end when range changes."""
        if scroll:
            self.rangeChanged.connect(self.on_scrollbar_range_changed)
        else:
            self.rangeChanged.disconnect(self.on_scrollbar_range_changed)


class AbstractSlider(AbstractSliderMixin, widgets.QAbstractSlider):
    """Integer value within a range."""


if __name__ == "__main__":
    app = widgets.app()
    slider = AbstractSlider()
    slider.setRange(0, 100)
    slider.show()
    app.exec()



from __future__ import annotations

import functools


def bold(text: str) -> str:
    return f"<b>{text}</b>"


def colored(text: str, color: str) -> str:
    return f"<font color={color!r}>{text}</font>"


@functools.cache
def color_text(input_text: str, text: str, color: str, case_sensitive: bool = False):
    """Color first occurences of input_text chars in text with given color."""

    def converter(x):
        return x if case_sensitive else x.lower()

    input_text = input_text if case_sensitive else input_text.lower()
    output_text = ""
    to_color = ""
    for char in text:
        if input_text and converter(char) == input_text[0]:
            to_color += char
            input_text = input_text[1:]
        else:
            if to_color:
                output_text += bold(colored(to_color, color))
                to_color = ""
            output_text += char
    if to_color:
        output_text += bold(colored(to_color, color))
    return output_text


if __name__ == "__main__":
    pat = "aab"
    candidates = ["aaaaab", "aacb", "abc", "abbaab"] * 10000
    print(color_text("ac", "abbbbc", "green"))



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtNetwork
from prettyqt.utils import bidict


mod = QtNetwork.QLocalSocket

LocalSocketErrorStr = Literal[
    "connection_refused",
    "peer_closed",
    "server_not_found",
    "socket_access",
    "socket_resource",
    "socket_timeout",
    "datagram_too_large",
    "connection",
    "unsupported_socket_operation",
    "operation",
    "unknown_socket",
]

LOCAL_SOCKET_ERROR: bidict[LocalSocketErrorStr, mod.LocalSocketError] = bidict(
    connection_refused=mod.LocalSocketError.ConnectionRefusedError,
    peer_closed=mod.LocalSocketError.PeerClosedError,
    server_not_found=mod.LocalSocketError.ServerNotFoundError,
    socket_access=mod.LocalSocketError.SocketAccessError,
    socket_resource=mod.LocalSocketError.SocketResourceError,
    socket_timeout=mod.LocalSocketError.SocketTimeoutError,
    datagram_too_large=mod.LocalSocketError.DatagramTooLargeError,
    connection=mod.LocalSocketError.ConnectionError,
    unsupported_socket_operation=mod.LocalSocketError.UnsupportedSocketOperationError,
    operation=mod.LocalSocketError.OperationError,
    unknown_socket=mod.LocalSocketError.UnknownSocketError,
)

LocalSocketStateStr = Literal[
    "unconnected",
    "connecting",
    "connected",
    "closing",
]

LOCAL_SOCKET_STATE: bidict[LocalSocketStateStr, mod.LocalSocketState] = bidict(
    unconnected=mod.LocalSocketState.UnconnectedState,
    connecting=mod.LocalSocketState.ConnectingState,
    connected=mod.LocalSocketState.ConnectedState,
    closing=mod.LocalSocketState.ClosingState,
)


class LocalSocket(core.IODeviceMixin, QtNetwork.QLocalSocket):
    """Local socket."""

    def __bool__(self):
        return self.isValid()

    def get_error(self) -> LocalSocketErrorStr:
        return LOCAL_SOCKET_ERROR.inverse[self.error()]

    def get_state(self) -> LocalSocketStateStr:
        return LOCAL_SOCKET_STATE.inverse[self.state()]

    def connect_to_server(
        self,
        server_name: str | None = None,
        mode: core.iodevice.OpenModeStr | core.QIODevice.OpenModeFlag = "read_write",
    ):
        open_mode = core.iodevice.OPEN_MODES.get_enum_value(mode)
        if server_name is not None:
            self.connectToServer(server_name, open_mode)
        else:
            self.connectToServer(open_mode)


if __name__ == "__main__":
    server = LocalSocket()



from __future__ import annotations

import emoji

from prettyqt import constants, core, gui


ZERO_COORD = core.Point(0, 0)


class EmojiIconEngine(gui.IconEngine):
    """A custom QIconEngine that can render emojis."""

    def __init__(self, name: str, language: str = "alias") -> None:
        self._name = name
        self._language = language
        super().__init__()

    def paint(
        self,
        painter: gui.QPainter,
        rect: core.QRect,
        mode: gui.QIcon.Mode,
        state: gui.QIcon.State,
    ):
        icon_char = emoji.emojize(f":{self._name}:", language=self._language)
        painter.drawText(rect, int(constants.ALIGN_CENTER), icon_char)

    def clone(self):
        """Required to subclass abstract QIconEngine."""
        return EmojiIconEngine(self._name, self._language)

    def pixmap(self, size, mode, state) -> gui.QPixmap:
        pm = gui.QPixmap(size)
        pm.fill(constants.GlobalColor.transparent)
        rect = core.Rect(ZERO_COORD, size)
        with gui.Painter(pm) as painter:
            self.paint(painter, rect, mode, state)
        return pm


class Emojis:
    def __init__(self):
        self._names = {v["en"][1:-1]: k for k, v in emoji.EMOJI_DATA.items()}

    def __getattr__(self, attribute):
        return self._names[attribute]

    def __dir__(self):
        return self._names


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    provider = Emojis()
    print(dir(provider))
    engine = EmojiIconEngine("thumbsup")
    icon = gui.QIcon(engine)
    widget = widgets.Widget()
    widget.set_icon(icon)
    widget.show()
    app.exec()



"""Provides QtStateMachine classes and functions."""

from prettyqt.qt import PYSIDE6


# elif PYQT6:
#     from PyQt6.QtStateMachine import *
if PYSIDE6:
    from PySide6.QtStateMachine import *
else:
    raise ModuleNotFoundError("No Qt bindings could be found")



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


class HelpSearchResult(QtHelp.QHelpSearchResult):
    """The data associated with the search result."""

    def get_url(self) -> core.Url:
        return core.Url(self.url())



from __future__ import annotations

import enum
import re
from typing import TYPE_CHECKING, Literal

import sublime_search

from prettyqt import constants, core
from prettyqt.utils import bidict, datatypes


if TYPE_CHECKING:
    from collections.abc import Iterable


class FilterMode(enum.IntEnum):
    fixed_string = 1
    fuzzy = 2
    wildcard = 4
    regex = 8


FilterModeStr = Literal["fixed_string", "fuzzy", "wildcard", "regex"]


FILTER_MODE: bidict[FilterModeStr, FilterMode] = bidict(
    fixed_string=FilterMode.fixed_string,
    fuzzy=FilterMode.fuzzy,
    wildcard=FilterMode.wildcard,
    regex=FilterMode.regex,
)


class SortFilterProxyModel(core.AbstractProxyModelMixin, core.QSortFilterProxyModel):
    """Support for sorting and filtering data passed between another model and a view."""

    FilterMode = core.Enum(FilterMode)
    invalidated = core.Signal()
    filter_mode_changed = core.Signal(str)
    ID = "sort_filter"
    ICON = "mdi.table-filter"

    def __init__(self, *args, **kwargs):
        self._filter_mode: FilterModeStr = "wildcard"
        super().__init__(*args, **kwargs)

    #     self._filter_column = 0

    # def setFilterKeyColumn(self, column: int | list[int] | None):
    #     if column is None:
    #         column = -1
    #     self._filter_column = column

    # def filterKeyColumn(self) -> int | list[int] | None:
    #     return self._filter_column

    # def filterAcceptsRow(self, source_row: int, source_index: core.ModelIndex):
    #     column = self.filterKeyColumn()
    #     col_count = self.sourceModel().columnCount()
    #     indexes = [
    #         self.sourceModel().index(source_row, i, source_index) for i in col_count
    #     ]
    #     labels = [self.sourceModel().data(idx) for idx in indexes]

    #     if isinstance(column, int) and source_index.column() == column:
    #         return super().filterAcceptsRow(source_row, source_index)

    def filterAcceptsRow(self, source_row: int, source_index: core.ModelIndex) -> bool:
        if self._filter_mode != "fuzzy":
            return super().filterAcceptsRow(source_row, source_index)
        column = self.filterKeyColumn()
        source_model = self.sourceModel()
        idx = source_model.index(source_row, column, source_index)
        text = source_model.data(idx)
        return sublime_search.fuzzy_match_simple(
            self.filterRegularExpression().pattern(),
            text,
            case_sensitive=self.is_filter_case_sensitive(),
        )

    def invalidate(self):
        super().invalidate()
        self.invalidated.emit()

    def lessThan(self, left, right) -> bool:
        role = super().sortRole()
        left_data = left.data(role)
        right_data = right.data(role)
        if left_data is not None and right_data is not None:
            return left_data < right_data
        return True

    def sort(
        self,
        column: int | None,
        ascending: bool | constants.SortOrder = constants.ASCENDING,
    ):
        if isinstance(ascending, bool):
            ascending = constants.ASCENDING if ascending else constants.DESCENDING
        if column is None:
            column = -1
        super().sort(column, ascending)

    def setFilterRegularExpression(self, pattern: datatypes.PatternType):
        if isinstance(pattern, re.Pattern):
            pattern = core.RegularExpression(pattern)
        super().setFilterRegularExpression(pattern)

    def setFilterString(self, search_str: str):
        pat = ".*?".join(map(re.escape, search_str))
        pat = f"(?=({pat}))"
        super().setFilterRegularExpression(pat)

    def setFixedFilterList(self, filter_list: Iterable[str]):
        pat = "|".join(filter_list)
        super().setFilterRegularExpression(pat)

    def set_filter_case_sensitivity(
        self, sensitivity: constants.CaseSensitivityStr | constants.CaseSensitivity
    ):
        """Set the filter case sensitivity.

        Args:
            sensitivity: filter case sensitivity
        """
        value = constants.CASE_SENSITIVITY.get_enum_value(sensitivity)
        super().setFilterCaseSensitivity(value)

    def get_filter_case_sensitivity(self) -> constants.CaseSensitivityStr:
        """Return current filter case sensitivity.

        Returns:
            filter case sensitivity
        """
        return constants.CASE_SENSITIVITY.inverse[super().filterCaseSensitivity()]

    def set_filter_case_sensitive(self, state: bool):
        if state:
            sensitivity = constants.CaseSensitivity.CaseSensitive
        else:
            sensitivity = constants.CaseSensitivity.CaseInsensitive
        super().setFilterCaseSensitivity(sensitivity)

    def is_filter_case_sensitive(self) -> bool:
        return super().filterCaseSensitivity() == constants.CaseSensitivity.CaseSensitive

    def set_sort_case_sensitive(self, state: bool):
        if state:
            sensitivity = constants.CaseSensitivity.CaseSensitive
        else:
            sensitivity = constants.CaseSensitivity.CaseInsensitive
        super().setSortCaseSensitivity(sensitivity)

    def is_sort_case_sensitive(self) -> bool:
        return super().sortCaseSensitivity() == constants.CaseSensitivity.CaseSensitive

    def get_filter_regular_expression(self) -> core.RegularExpression:
        return core.RegularExpression(self.filterRegularExpression())

    def set_sort_role(self, role: constants.ItemDataRoleStr | int):
        role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        super().setSortRole(role)

    def set_filter_role(self, role: constants.ItemDataRoleStr | int):
        role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        super().setFilterRole(role)

    def get_sort_order(self) -> Literal["ascending", "descending"]:
        return "ascending" if super().sortOrder() == constants.ASCENDING else "descending"

    def set_filter_key_column(self, column: int | None):
        if column is None:
            column = -1
        super().setFilterKeyColumn(column)

    def set_search_term(self, search_term: str | Iterable[str]):
        match self._filter_mode:
            case "fixed_string" if isinstance(search_term, list):
                self.setFixedFilterList(search_term)
            case "fixed_string":
                self.setFilterFixedString(search_term)
            case "substring":
                self.setFilterString(search_term)
            case "fuzzy":
                self.setFilterFixedString(search_term)
            case "wildcard":
                self.setFilterWildcard(search_term)
            case "regex":
                self.setFilterRegularExpression(search_term)

    def get_filter_mode(self) -> FilterModeStr:
        return self._filter_mode

    def set_filter_mode(self, mode: FilterModeStr):
        self._filter_mode = mode
        self.filter_mode_changed.emit(mode)

    filterMode = core.Property(  # noqa: N815
        str,
        get_filter_mode,
        set_filter_mode,
        notify=filter_mode_changed,
        doc="Mode to use for filtering",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    from prettyqt.itemmodels.importlibdistributionmodel import (
        ImportlibDistributionModel,
    )

    source_model = ImportlibDistributionModel.from_package("prettyqt")
    model = SortFilterProxyModel()
    model.setSourceModel(source_model)
    model.set_filter_mode("fuzzy")
    lineedit = widgets.LineEdit()
    lineedit.value_changed.connect(model.set_search_term)
    widget = widgets.Widget()
    widget.set_layout("vertical")
    widget.box.add(lineedit)
    table = widgets.TableView()
    table.setSortingEnabled(True)
    widget.box.add(table)
    table.set_model(model)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core
from prettyqt.qt import QtLocation


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class PlaceIcon(QtLocation.QPlaceIcon):
    """Represents an icon."""

    def __bool__(self):
        return not self.isEmpty()

    def __setitem__(self, index: str, val: datatypes.Variant):
        attrs = self.parameters()
        attrs[index] = val
        self.setParameters(attrs)

    def __getitem__(self, index: str) -> datatypes.Variant:
        attr = self.parameters()
        if index not in attr:
            msg = f"Key {index!r} does not exist."
            raise KeyError(msg)
        return attr[index]

    # def get_manager(self) -> location.PlaceManager:
    #     return location.PlaceManager(self.manager())

    def get_url(self) -> core.Url:
        return core.Url(self.url())


if __name__ == "__main__":
    icon = PlaceIcon()
    print(bool(icon))



from __future__ import annotations

import pathlib

from prettyqt import core, qml


class QmlExpression(core.ObjectMixin, qml.QQmlExpression):
    """Evaluates JavaScript in a QML context."""

    def get_source_file(self) -> pathlib.Path | None:
        return pathlib.Path(source) if (source := self.sourceFile()) else None

    def get_error(self) -> qml.QmlError:
        return qml.QmlError(self.error())


if __name__ == "__main__":
    exp = QmlExpression()
    print(exp.get_error())



"""Provides a menu that display the list of recent files and a RecentFilesManager.

Application's QSettings are used to store the list of recent files.
"""

from __future__ import annotations

import contextlib
import os
import pathlib
import sys

from prettyqt import core, gui, iconprovider, widgets


def _(text):
    return text


class RecentFilesManager(core.Object):
    """Manages a list of recent files. The list of files is stored in QSettings."""

    #: Maximum number of files kept in the list.
    max_recent_files = 15
    updated = core.Signal()

    def __init__(self, organisation, application, qsettings=None):
        super().__init__()
        if qsettings is None:
            self._settings = core.Settings(organisation, application)
        else:
            self._settings = qsettings

    def clear(self):
        """Clears recent files in QSettings."""
        self.set_value("list", [])
        self.updated.emit()

    def remove(self, filename: os.PathLike):
        """Remove a file path from the list of recent files.

        :param filename: Path of the file to remove
        """
        files = self.get_value("list", [])
        files.remove(os.fspath(filename))
        self.set_value("list", files)
        self.updated.emit()

    def get_value(self, key, default=None):
        """Reads value from QSettings.

        :param key: value key
        :param default: default value.
        :return: value
        """

        def unique(seq, idfun=None):
            if idfun is None:

                def idfun(x):
                    return x

            # order preserving
            seen = {}
            result = []
            for item in seq:
                marker = idfun(item)
                if marker in seen:
                    continue
                seen[marker] = 1
                result.append(item)
            return result

        val = self._settings.value(f"recent_files/{key}", default)
        match val:
            case None:
                lst: list[str] = []
            case str():
                lst = [val]
            case _:
                lst = val
        return unique([os.path.normpath(pth) for pth in lst])

    def set_value(self, key, value):
        """Set the recent files value in QSettings.

        :param key: value key
        :param value: new value
        """
        if value is None:
            value = []
        value = [os.path.normpath(pth) for pth in value]
        self._settings.setValue(f"recent_files/{key}", value)

    def get_recent_files(self) -> list[str]:
        """Gets the list of recent files.

        Files that do not exists anymore are automatically filtered
        """
        ret_val = []
        files = self.get_value("list", [])
        # filter files, remove files that do not exist anymore
        for file in files:
            if file is not None and pathlib.Path(file).exists():
                if (
                    os.path.ismount(file)
                    and sys.platform == "win32"
                    and not file.endswith("\\")
                ):
                    file += "\\"
                if file not in ret_val:
                    ret_val.append(file)
        return ret_val

    def open_file(self, file):
        """Adds a file to the list (and move it to top if the file already exists).

        :param file: file path to add the list of recent files.
        """
        files = self.get_recent_files()
        with contextlib.suppress(ValueError):
            files.remove(file)

        files.insert(0, file)
        # discard old files
        del files[self.max_recent_files :]
        self.set_value("list", files)
        self.updated.emit()

    def last_file(self):
        """Returns the path to the last opened file."""
        files = self.get_recent_files()
        try:
            return files[0]
        except IndexError:
            return None


class MenuRecentFiles(widgets.Menu):
    """Menu that manage the list of recent files.

    To use the menu, simply connect to the open_requested signal.

    """

    #: Signal emitted when the user clicked on a recent file action.
    #: The parameter is the path of the file to open.
    open_requested = core.Signal(str)
    clear_requested = core.Signal()

    def __init__(
        self,
        parent,
        recent_files_manager=None,
        title: str = "Recent files",
        icon_provider=None,
        clear_icon=None,
    ):
        """Init.

        :param parent: parent object
        :param icon_provider: Object that provides icon based on the file path.
        :type icon_provider: QtWidgets.QFileIconProvider
        :param clear_icon: Clear action icon. This parameter is a tuple made up
            of the icon theme name and the fallback icon path (from your
            resources). Default is None, clear action has no icons.
        """
        super().__init__(title, parent)
        if icon_provider is None:
            self.icon_provider = widgets.FileIconProvider()
        else:
            self.icon_provider = icon_provider
        self.clear_icon = clear_icon
        #: Recent files manager
        self.manager = recent_files_manager
        #: List of recent files actions
        self.recent_files_actions: list[gui.Action] = []
        self.update_actions()

    def update_actions(self):
        """Updates the list of actions."""
        self.clear()
        self.recent_files_actions[:] = []
        for file in self.manager.get_recent_files():
            icon = self.icon_provider.icon(core.FileInfo(file))
            action = gui.Action(
                parent=self,
                text=os.path.split(file)[1],
                tool_tip=file,
                status_tip=file,
                icon=icon,
                triggered=self._on_action_triggered,
            )
            action.setData(file)
            self.addAction(action)
            self.recent_files_actions.append(action)
        self.addSeparator()
        action_clear = gui.Action(text="Clear list", triggered=self.clear_recent_files)
        action_clear.setIcon(iconprovider.get_icon("fa.times-circle"))
        self.addAction(action_clear)

    def clear_recent_files(self):
        """Clear recent files and menu."""
        self.manager.clear()
        self.update_actions()
        self.clear_requested.emit()

    def _on_action_triggered(self):
        """Emits open_requested when a recent file action has been triggered."""
        action = self.sender()
        assert isinstance(action, gui.Action)
        path = action.data()
        self.open_requested.emit(path)
        self.update_actions()



"""Provides Qt init stuff."""

import os
import sys

from typing import Literal
import warnings


API_NAMES = {"pyqt6": "PyQt6", "pyside6": "PySide6"}


DEFAULT = "pyqt6"

# Detecting if a binding was specified by the user
binding_specified = "QT_API" in os.environ
API = os.environ.get("QT_API", DEFAULT).lower()

if API.lower() in {"pyqt5", "pyside2"}:
    API = DEFAULT
    binding_specified = False

initial_api = API
if API not in API_NAMES:
    msg = f"Specified QT_API={repr(API)!r} is not in valid options: {API_NAMES}"
    raise ValueError(msg)


PYQT6 = False
PYSIDE6 = False

API_VERSION = None
QT_VERSION = None


if "PyQt6" in sys.modules:
    API = "pyqt6"
elif "PySide6" in sys.modules:
    API = "pyside6"

if API == "pyqt6":
    try:
        from PyQt6.QtCore import (
            PYQT_VERSION_STR as API_VERSION,  # analysis:ignore
            QT_VERSION_STR as QT_VERSION,  # analysis:ignore
        )

        PYQT6 = True

    except ImportError:
        API = "pyside6"
    # else:
    #     os.environ["QT_API"] = API

if API == "pyside6":
    try:
        from PySide6 import (
            __version__ as API_VERSION,  # noqa: F401, N812
        )  # analysis:ignore
        from PySide6.QtCore import (
            __version__ as QT_VERSION,  # noqa: F401, N812
        )  # analysis:ignore

        PYSIDE6 = True

    except ImportError as e:
        raise ImportError(API) from e
    # else:
    #     os.environ["QT_API"] = API


# If a correct API name is passed to QT_API and it could not be found,
# switches to another and informs through the warning
if initial_api != API and binding_specified:
    warnings.warn(
        f"Selected binding {initial_api!r} not found. Falling back to {API!r}",
        ImportWarning,
        stacklevel=2,
    )


# Set display name of the Qt API
API_NAME = API_NAMES[API]
MODULE = sys.modules[API_NAME]


def set_env_vars(qt_binding: Literal["PyQt6", "PySide6"]):
    env_vars = ["QT_API", "USE_QT_API", "PYTEST_QT_API", "PYQTGRAPH_QT_LIB"]
    for var in env_vars:
        os.environ[var] = qt_binding


set_env_vars(API_NAME)


# from prettyqt.qt import QtCore, QtGui, QtWidgets
# from prettyqt import paths
# import inspect
# import pathlib

# module_dict = dict(QtWidgets=QtWidgets, QtGui=QtGui, QtCore=QtCore)

# for module_name, module in module_dict.items():
#     clsmembers = inspect.getmembers(module, inspect.isclass)
#     for klass_name, klass in clsmembers:
#         path = paths.DOCSTRING_PATH /  module_name
#         filepath = path / f"{klass_name}.txt"
#         if filepath.exists():
#             # print(klass_name)
#             text = filepath.read_text()
#             klass.__doc__ = text



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.utils import bidict


DiscoveryMethodStr = Literal["none", "classic", "low_energy"]

DISCOVERY_METHODS: bidict[
    DiscoveryMethodStr, bluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod
] = bidict(
    none=bluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.NoMethod,
    classic=bluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.ClassicMethod,
    low_energy=bluetooth.QBluetoothDeviceDiscoveryAgent.DiscoveryMethod.LowEnergyMethod,
)

Error = bluetooth.QBluetoothDeviceDiscoveryAgent.Error

ErrorStr = Literal[
    "none",
    "powered_off",
    "input_output",
    "invalid_bluetooth_adapter",
    "unsupported_platform",
    "unsupported_discovery",
    "unknown",
]

ERROR: bidict[ErrorStr, Error] = bidict(
    none=Error.NoError,
    powered_off=Error.PoweredOffError,
    input_output=Error.InputOutputError,
    invalid_bluetooth_adapter=Error.InvalidBluetoothAdapterError,
    unsupported_platform=Error.UnsupportedPlatformError,
    unsupported_discovery=Error.UnsupportedDiscoveryMethod,
    unknown=Error.UnknownError,
)

# INQUIRY_TYPES = bidict(
#     unlimited=bluetooth.QBluetoothDeviceDiscoveryAgent.GeneralUnlimitedInquiry,
#     limited=bluetooth.QBluetoothDeviceDiscoveryAgent.LimitedInquiry,
# )

# InquiryTypeStr = Literal["unlimited", "limited"]


class BluetoothDeviceDiscoveryAgent(
    core.ObjectMixin, bluetooth.QBluetoothDeviceDiscoveryAgent
):
    """Discovers the Bluetooth devices nearby."""

    device_discovered = core.Signal(object)  # bluetooth.BluetoothDeviceInfo casts signal.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deviceDiscovered.connect(self._on_device_discovered)

    def _on_device_discovered(self, info):
        info = bluetooth.BluetoothDeviceInfo(info)
        self.device_discovered.emit(info)

    # def set_inquiry_type(self, typ: InquiryTypeStr):
    #     """Set inquiry type.

    #     Args:
    #         typ: inquiry type
    #     """
    #     self.setInquiryType(INQUIRY_TYPES.get_enum_value(typ))

    # def get_inquiry_type(self) -> InquiryTypeStr:
    #     """Get the current inquiry type.

    #     Returns:
    #         inquiry type
    #     """
    #     return INQUIRY_TYPES.inverse[self.inquiryType()]

    def start_discovery(self, classic: bool = True, low_energy: bool = True):
        """Start discovery of bluetooth devices.

        Arguments:
            classic: use classic method
            low_energy: use BLE mode
        """
        flag = bluetooth.QBluetoothDeviceDiscoveryAgent.NoMethod
        if classic:
            flag |= bluetooth.QBluetoothDeviceDiscoveryAgent.ClassicMethod
        if low_energy:
            flag |= bluetooth.QBluetoothDeviceDiscoveryAgent.LowEnergyMethod
        self.start(flag)

    def get_error(self) -> ErrorStr:
        """Get error type if available."""
        return ERROR.inverse[self.error()]

    def get_supported_discovery_methods(self) -> list[DiscoveryMethodStr]:
        """Return a list of supported discovery methods."""
        return DISCOVERY_METHODS.get_list(self.supportedDiscoveryMethods())


if __name__ == "__main__":
    import logging

    from prettyqt import core

    app = core.app()
    agent = BluetoothDeviceDiscoveryAgent(app)
    agent.setLowEnergyDiscoveryTimeout(500)
    agent.device_discovered.connect(lambda x: logging.info(x))
    agent.finished.connect(app.quit)
    agent.start_discovery()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import ctypes
import math
from typing import TYPE_CHECKING, Self

from prettyqt import constants, core, gui
from prettyqt.qt import API
from prettyqt.utils import datatypes, serializemixin


if TYPE_CHECKING:
    from collections.abc import Iterator


class PolygonF(serializemixin.SerializeMixin, gui.QPolygonF):
    """List of points using floating point precision."""

    def __repr__(self):
        return f"{type(self).__name__}(<{len(self)} points>)"

    def __iter__(self) -> Iterator[core.PointF]:
        return iter(self.get_point(i) for i in range(self.size()))

    def __len__(self) -> int:
        return self.size()

    def __bool__(self):
        return not self.isEmpty()

    def __contains__(self, point: datatypes.PointFType) -> bool:
        if isinstance(point, tuple):
            point = core.PointF(*point)
        return self.containsPoint(point, constants.FillRule.OddEvenFill)

    def __getitem__(self, index: int) -> core.PointF:
        if index >= self.size():
            raise KeyError(index)
        return self.get_point(index)

    # def __setitem__(self, index: int, value: datatypes.PointType):
    #     if isinstance(value, tuple):
    #         self.setPoint(index, *value)
    #     else:
    #         self.setPoint(index, value)

    def __sub__(self, other: gui.QPolygonF) -> Self:
        return type(self)(self.subtracted(other))

    def __and__(self, other: gui.QPolygonF) -> Self:  # &
        return type(self)(self.intersected(other))

    def __xor__(self, other: gui.QPolygonF) -> Self:  # ^
        union = self | other
        intersect = self & other
        return union - intersect

    def __or__(self, other: gui.QPolygonF) -> Self:  # |
        return type(self)(self.united(other))

    def __eq__(self, other: object) -> bool:
        return (
            list(self) == [other.at(i) for i in range(other.size())]
            if isinstance(other, type(self))
            else False
        )

    def get_point(self, index: int) -> core.PointF:
        return core.PointF(self.at(index))

    def get_points(self) -> list[core.PointF]:
        return [self.get_point(i) for i in range(self.size())]

    def add_points(self, *points: datatypes.PointFType):
        for p in points:
            self.append(datatypes.to_pointf(p))

    def to_polygon(self) -> gui.Polygon:
        return gui.Polygon(self.toPolygon())

    @classmethod
    def create_star(cls, scale: int = 1) -> Self:
        poly = cls()
        poly.append(core.PointF(0.5 * scale, 0.0))
        for i in range(1, 5):
            val = 0.8 * i * math.pi
            point = core.PointF(0.5 * scale * math.cos(val), 0.5 * scale * math.sin(val))
            poly.append(point)
        return poly

    @classmethod
    def create_diamond(cls, scale: int = 1) -> Self:
        points = [
            core.PointF(-0.1 * scale, 0.0),
            core.PointF(0.0, -0.1 * scale),
            core.PointF(0.1 * scale, 0.0),
            core.PointF(0.0, 0.1 * scale),
            core.PointF(-0.1 * scale, 0.0),
        ]
        poly = gui.PolygonF()
        poly.add_points(*points)
        return poly

    def get_data_buffer(self, size: int):
        self.resize(size)
        if API == "pyside6":
            import shiboken6

            address = shiboken6.getCppPointer(self.data())
            buffer = (ctypes.c_double * 2 * self.size()).from_address(address[0])
        else:
            buffer = self.data()
            buffer.setsize(16 * self.size())
        return buffer

    @classmethod
    def from_xy(cls, xdata, ydata) -> Self:
        import numpy as np

        size = len(xdata)
        polyline = cls()
        buffer = polyline.get_data_buffer(size)
        memory = np.frombuffer(buffer, np.float64)
        memory[: (size - 1) * 2 + 1 : 2] = np.array(xdata, dtype=np.float64, copy=False)
        memory[1 : (size - 1) * 2 + 2 : 2] = np.array(ydata, dtype=np.float64, copy=False)
        return polyline


if __name__ == "__main__":
    poly = PolygonF([core.Point(1, 1), core.Point(2, 2)])
    poly2 = PolygonF([core.Point(1, 1), core.Point(2, 2)])
    new = poly | poly2
    print(repr(new))



from __future__ import annotations

import ctypes
from typing import TYPE_CHECKING, Self

from prettyqt import constants, core
from prettyqt.qt import API, QtGui
from prettyqt.utils import datatypes, serializemixin


if TYPE_CHECKING:
    from collections.abc import Iterator


class Polygon(serializemixin.SerializeMixin, QtGui.QPolygon):
    """List of points using integer precision."""

    def __repr__(self):
        return f"{type(self).__name__}(<{len(self)} points>)"

    def __iter__(self) -> Iterator[core.Point]:
        return iter(self.get_point(i) for i in range(self.size()))

    def __len__(self) -> int:
        return self.size()

    def __bool__(self):
        return not self.isEmpty()

    def __contains__(self, point: datatypes.PointType) -> bool:
        if isinstance(point, tuple):
            point = core.Point(*point)
        return self.containsPoint(point, constants.FillRule.OddEvenFill)

    def __getitem__(self, index: int) -> core.Point:
        if index >= self.size():
            raise KeyError(index)
        return self.get_point(index)

    def __setitem__(self, index: int, value: datatypes.PointType):
        p = core.Point(*value) if isinstance(value, tuple) else value
        # PySide6 workaround: setPoint does not exist
        self.remove(index)
        self.insert(index, p)

    def __sub__(self, other: QtGui.QPolygon) -> Polygon:
        return Polygon(self.subtracted(other))

    def __and__(self, other: QtGui.QPolygon) -> Polygon:  # &
        return Polygon(self.intersected(other))

    def __xor__(self, other: QtGui.QPolygon) -> Polygon:  # ^
        union = self | other
        intersect = self & other
        return union - intersect

    def __or__(self, other: QtGui.QPolygon) -> Polygon:  # |
        return Polygon(self.united(other))

    def get_point(self, index: int) -> core.Point:
        # PySide6 doesnt have self.point method
        return core.Point(self.value(index))

    def get_points(self) -> list[core.Point]:
        return [self.get_point(i) for i in range(self.size())]

    def add_points(self, *points: datatypes.PointType):
        for p in points:
            point = core.Point(*p) if isinstance(p, tuple) else p
            self.append(point)

    def get_data_buffer(self, size: int):
        self.resize(size)
        if API == "pyside6":
            import shiboken6

            address = shiboken6.getCppPointer(self.data())
            buffer = (ctypes.c_long * 2 * self.size()).from_address(address[0])
        else:
            buffer = self.data()
            buffer.setsize(8 * self.size())
        return buffer

    @classmethod
    def from_xy(cls, xdata, ydata) -> Self:
        import numpy as np

        size = len(xdata)
        polyline = cls()
        buffer = polyline.get_data_buffer(size)
        memory = np.frombuffer(buffer, np.float64)
        memory[: (size - 1) * 2 + 1 : 2] = np.array(xdata, dtype=np.float64, copy=False)
        memory[1 : (size - 1) * 2 + 2 : 2] = np.array(ydata, dtype=np.float64, copy=False)
        return polyline


if __name__ == "__main__":
    poly = Polygon([core.Point(1, 1), core.Point(2, 2)])
    poly2 = Polygon([core.Point(1, 1), core.Point(2, 2)])
    poly.get_data_buffer(10)
    poly & poly2



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui
from prettyqt.utils import bidict


LineHeightTypeStr = Literal[
    "single",
    "proportional",
    "fixed",
    "minimum",
    "line_distance",
]

LINE_HEIGHT_TYPES: bidict[MarkerTypeStr, gui.QTextBlockFormat.LineHeightTypes] = bidict(
    single=gui.QTextBlockFormat.LineHeightTypes.SingleHeight,
    proportional=gui.QTextBlockFormat.LineHeightTypes.ProportionalHeight,
    fixed=gui.QTextBlockFormat.LineHeightTypes.FixedHeight,
    minimum=gui.QTextBlockFormat.LineHeightTypes.MinimumHeight,
    line_distance=gui.QTextBlockFormat.LineHeightTypes.LineDistanceHeight,
)

MarkerTypeStr = Literal[
    "none",
    "unchecked",
    "checked",
]

MARKER_TYPE: bidict[MarkerTypeStr, gui.QTextBlockFormat.MarkerType] = bidict(
    none=gui.QTextBlockFormat.MarkerType.NoMarker,
    unchecked=gui.QTextBlockFormat.MarkerType.Unchecked,
    checked=gui.QTextBlockFormat.MarkerType.Checked,
)


class TextBlockFormat(gui.TextFormatMixin, gui.QTextBlockFormat):
    def set_marker(self, marker: MarkerTypeStr | gui.QTextBlockFormat.MarkerType):
        """Set the marker.

        Args:
            marker: marker
        """
        self.setMarker(MARKER_TYPE.get_enum_value(marker))

    def get_marker(self) -> MarkerTypeStr:
        """Get current marker.

        Returns:
            current marker
        """
        return MARKER_TYPE.inverse[self.marker()]

    def set_alignment(self, alignment: constants.AlignmentStr | constants.AlignmentFlag):
        """Set the alignment of the format.

        Args:
            alignment: alignment for the format
        """
        self.setAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def set_page_break_policy(
        self, policy: gui.textformat.PageBreakFlagStr | gui.QTextFormat.PageBreakFlag
    ):
        """Set page break policy.

        Args:
            policy: page break policy
        """
        self.setPageBreakPolicy(gui.textformat.PAGE_BREAK_FLAG.get_enum_value(policy))

    def get_page_break_policy(self) -> gui.textformat.PageBreakFlagStr:
        """Get the current page break policy.

        Returns:
            page break policy
        """
        return gui.textformat.PAGE_BREAK_FLAG.inverse[self.pageBreakPolicy()]



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtPositioning
from prettyqt.utils import bidict, get_repr


mod = QtPositioning.QGeoCoordinate

COORDINATE_FORMAT = bidict(
    default=mod.CoordinateFormat.Degrees,
    with_hemisphere=mod.CoordinateFormat.DegreesWithHemisphere,
    min=mod.CoordinateFormat.DegreesMinutes,
    min_with_hemisphere=mod.CoordinateFormat.DegreesMinutesWithHemisphere,
    min_sec=mod.CoordinateFormat.DegreesMinutesSeconds,
    min_sec_with_hemisphere=mod.CoordinateFormat.DegreesMinutesSecondsWithHemisphere,
)

CoordinateFormatStr = Literal[
    "default",
    "with_hemisphere",
    "min",
    "min_with_hemisphere",
    "min_sec",
    "min_sec_with_hemisphere",
]


COORDINATE_TYPE = bidict({
    "invalid": mod.CoordinateType.InvalidCoordinate,
    "2d": mod.CoordinateType.Coordinate2D,
    "3d": mod.CoordinateType.Coordinate3D,
})

CoordinateTypeStr = Literal["invalid", "2d", "3d"]


class GeoCoordinate(QtPositioning.QGeoCoordinate):
    def __str__(self):
        return self.toString()

    def __format__(self, format_spec: CoordinateFormatStr):
        return self.toString(COORDINATE_FORMAT[format_spec])

    def __repr__(self):
        return get_repr(self, self.latitude(), self.longitude())

    def __bool__(self):
        return self.isValid()

    def get_type(self) -> str:
        return COORDINATE_TYPE.inverse[self.type()]


if __name__ == "__main__":
    coord = GeoCoordinate(11, 11)
    print(coord)
    print(repr(coord))



from __future__ import annotations

from prettyqt import location


class PlaceResult(location.PlaceSearchResultMixin, location.QPlaceResult):
    """Represents a search result containing a place."""

    def get_place(self) -> location.Place:
        return location.Place(self.place())



from __future__ import annotations

from typing import TYPE_CHECKING, Self, overload

from prettyqt import constants, widgets
from prettyqt.utils import listdelegators


if TYPE_CHECKING:
    from collections.abc import Iterator


class SplitterMixin(widgets.FrameMixin):
    def __init__(
        self,
        orientation: constants.OrientationStr | constants.Orientation = "horizontal",
        **kwargs,
    ):
        super().__init__(constants.ORIENTATION.get_enum_value(orientation), **kwargs)
        self.setHandleWidth(10)

    @overload
    def __getitem__(self, index: int | str) -> widgets.QWidget: ...

    @overload
    def __getitem__(
        self, index: slice
    ) -> listdelegators.SplitterDelegator[widgets.QWidget]: ...

    def __getitem__(
        self, index: int | str | slice
    ) -> widgets.QWidget | listdelegators.SplitterDelegator[widgets.QWidget]:
        match index:
            case int():
                if index >= self.count():
                    raise IndexError(index)
                return self.widget(index)
            case str():
                result = self.find_child(widgets.QWidget, index)
                if result is None:
                    msg = "Widget not found"
                    raise KeyError(msg)
                return result
            case slice():
                stop = index.stop or self.count()
                rng = range(index.start or 0, stop, index.step or 1)
                wdgs = [self.widget(i) for i in rng]
                return listdelegators.SplitterDelegator(wdgs, parent=self)
            case _:
                raise TypeError(index)

    def __setitem__(self, index: int, value: widgets.QWidget):
        self.replaceWidget(index, value)

    def __iter__(self) -> Iterator[widgets.QWidget]:
        return iter(self.get_widgets())

    def __len__(self) -> int:
        return self.count()

    def __contains__(self, item: widgets.QWidget):
        return self.indexOf(item) >= 0

    def __add__(self, other: widgets.QWidget | widgets.QLayout):
        self.add(other)
        return self

    def __iadd__(self, item):
        self.add(item)
        return self

    def createHandle(self) -> widgets.SplitterHandle:
        return widgets.SplitterHandle(self.orientation(), self)

    # def saveState(self):
    #     sizes = self.sizes()
    #     if all(x == 0 for x in sizes):
    #         sizes = [10] * len(sizes)
    #     return {'sizes': sizes}

    # def restoreState(self, state):
    #     sizes = state['sizes']
    #     self.setSizes(sizes)
    #     for i in range(len(sizes)):
    #         self.setStretchFactor(i, sizes[i])

    def get_widgets(self) -> listdelegators.SplitterDelegator[widgets.QWidget]:
        widgets = [self.widget(i) for i in range(self.count())]
        return listdelegators.SplitterDelegator(widgets, parent=self)

    def add_widget(
        self,
        widget: widgets.QWidget,
        stretch: int | None = None,
        collapsible: bool = True,
        position: int | None = None,
    ) -> widgets.QWidget:
        if position is None:
            self.addWidget(widget)
        else:
            self.insertWidget(position, widget)
        index = self.indexOf(widget)
        if stretch is not None:
            self.setStretchFactor(index, stretch)
        self.setCollapsible(index, collapsible)
        return widget

    def add_layout(
        self,
        layout: widgets.QLayout,
        stretch: int | None = None,
        collapsible: bool = True,
        position: int | None = None,
    ) -> widgets.Widget:
        widget = widgets.Widget()
        widget.set_layout(layout)
        return self.add_widget(
            widget, stretch=stretch, collapsible=collapsible, position=position
        )

    def add(
        self,
        item: widgets.QWidget | widgets.QLayout | list[widgets.QWidget | widgets.QLayout],
        stretch: int | None = None,
    ):
        match item:
            case widgets.QWidget():
                self.add_widget(item, stretch=stretch)
            case widgets.QLayout():
                self.add_layout(item)
            case list():
                for i in item:
                    self.add(i)
        return item

    @classmethod
    def from_widgets(
        cls,
        *widgets: widgets.QWidget,
        horizontal: bool = False,
        parent: widgets.QWidget | None = None,
    ) -> Self:
        splitter = cls("horizontal" if horizontal else "vertical", parent=parent)
        for widget in widgets:
            splitter += widget
        return splitter

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the splitter.

        Args:
            orientation: orientation for the splitter

        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]


class Splitter(SplitterMixin, widgets.QSplitter):
    """Implements a splitter widget."""


if __name__ == "__main__":
    app = widgets.app()
    widget = Splitter()
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


mod = core.QCommandLineParser
op = mod.OptionsAfterPositionalArgumentsMode

OptionsAfterPosArgStr = Literal["options", "positional_arguments"]

OPTIONS_AFTER_POS_ARG: bidict[OptionsAfterPosArgStr, op] = bidict(
    options=op.ParseAsOptions,
    positional_arguments=op.ParseAsPositionalArguments,
)

SingleDashWordStr = Literal["compacted_short", "long"]

SINGLE_DASH_WORD: bidict[SingleDashWordStr, mod.SingleDashWordOptionMode] = bidict(
    compacted_short=mod.SingleDashWordOptionMode.ParseAsCompactedShortOptions,
    long=mod.SingleDashWordOptionMode.ParseAsLongOptions,
)


class CommandLineParser(core.QCommandLineParser):
    """Means for handling the command line options."""

    def set_single_dash_word_option_mode(self, mode: SingleDashWordStr):
        """Set the single dash word option mode.

        Args:
            mode: single dash word option mode
        """
        self.setSingleDashWordOptionMode(SINGLE_DASH_WORD.get_enum_value(mode))

    def set_options_after_positional_arguments_mode(
        self, mode: SingleDashWordStr | mod.SingleDashWordOptionMode
    ):
        """Set the options after positional arguments mode.

        Args:
            mode: options after positional arguments mode
        """
        val = OPTIONS_AFTER_POS_ARG.get_enum_value(mode)
        self.setOptionsAfterPositionalArgumentsMode(val)

    def add_option(
        self,
        name: str,
        description: str | None = None,
        value_name: str | None = None,
        default_value: str | None = None,
    ) -> core.CommandLineOption:
        if description is None:
            description = ""
        if value_name is None:
            value_name = ""
        if default_value is None:
            default_value = ""
        option = core.CommandLineOption(name, description, value_name, default_value)
        self.addOption(option)
        return option



from __future__ import annotations

import contextlib

from qstylizer import parser

from prettyqt import gui


class QssValidator(gui.Validator):
    """Validator which checks whether given string is a parseable css string."""

    ID = "qss"

    def __eq__(self, other: object):
        return isinstance(other, QssValidator)

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        with contextlib.suppress(ValueError):
            if not text or parser.parse(text):
                return self.State.Acceptable, text, pos
        return self.State.Intermediate, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = QssValidator()
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.setValidator(val)
    widget.show()
    with app.debug_mode():
        app.exec()
    print(widget.hasAcceptableInput())



"""Classes for unit testing Qt applications and libraries."""

from .abstractitemmodeltester import AbstractItemModelTester
from .signalspy import SignalSpy
from prettyqt.qt import QtTest

QT_MODULE = QtTest

__all__ = ["AbstractItemModelTester", "SignalSpy"]



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtQuick


class QuickTextDocument(core.ObjectMixin, QtQuick.QQuickTextDocument):
    """Access to the QTextDocument of QQuickTextEdit."""


if __name__ == "__main__":
    item = QtQuick.QQuickItem()
    doc = QuickTextDocument(item)



from __future__ import annotations

from typing import Literal, Self

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, datatypes


InputStr = Literal["press", "move", "release"]

INPUT: bidict[InputStr, widgets.QScroller.Input] = bidict(
    press=widgets.QScroller.Input.InputPress,
    move=widgets.QScroller.Input.InputMove,
    release=widgets.QScroller.Input.InputRelease,
)

ScrollGestureTypeStr = Literal[
    "touch", "left_mouse_button", "middle_mouse_button", "right_mouse_button"
]

SCROLLER_GESTURE_TYPE: bidict[
    ScrollGestureTypeStr, widgets.QScroller.ScrollerGestureType
] = bidict(
    touch=widgets.QScroller.ScrollerGestureType.TouchGesture,
    left_mouse_button=widgets.QScroller.ScrollerGestureType.LeftMouseButtonGesture,
    middle_mouse_button=widgets.QScroller.ScrollerGestureType.MiddleMouseButtonGesture,
    right_mouse_button=widgets.QScroller.ScrollerGestureType.RightMouseButtonGesture,
)


StateStr = Literal["inactive", "pressed", "dragging", "scrolling"]

STATE: bidict[StateStr, widgets.QScroller.State] = bidict(
    inactive=widgets.QScroller.State.Inactive,
    pressed=widgets.QScroller.State.Pressed,
    dragging=widgets.QScroller.State.Dragging,
    scrolling=widgets.QScroller.State.Scrolling,
)


class Scroller(core.ObjectMixin):
    """Enables kinetic scrolling for any scrolling widget or graphics item."""

    def __init__(self, item: widgets.QScroller):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_state(self) -> StateStr:
        """Return current state.

        Returns:
            state
        """
        return STATE.inverse[self.state()]

    def get_velocity(self) -> core.PointF:
        return core.PointF(self.velocity())

    def get_pixel_per_meter(self) -> core.PointF:
        return core.PointF(self.pixelPerMeter())

    def get_final_position(self) -> core.PointF:
        return core.PointF(self.finalPosition())

    def handle_input(
        self,
        input_type: InputStr | widgets.QScroller.Input,
        position: datatypes.PointFType,
        timestamp: int = 0,
    ) -> bool:
        return self.handleInput(
            INPUT.get_enum_value(input_type), datatypes.to_pointf(position), timestamp
        )

    def get_scroller_properties(self) -> widgets.ScrollerProperties:
        return widgets.ScrollerProperties(self.scrollerProperties())

    @classmethod
    def get_scroller(cls, obj: core.QObject) -> Self:
        return cls(widgets.QScroller.scroller(obj))

    @staticmethod
    def grab_gesture(
        target: core.QObject,
        gesture_type: (
            ScrollGestureTypeStr | widgets.QScroller.ScrollerGestureType
        ) = "touch",
    ) -> constants.GestureTypeStr:
        gesture = widgets.QScroller.grabGesture(
            target, SCROLLER_GESTURE_TYPE.get_enum_value(gesture_type)
        )
        if gesture >= 256:
            gesture -= 256
        return constants.GESTURE_TYPE.inverse[gesture]

    @staticmethod
    def grabbed_gesture(target: core.QObject) -> constants.GestureTypeStr:
        return constants.GESTURE_TYPE.inverse[widgets.QScroller.grabbedGesture(target)]


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.QPlainTextEdit()
    scroller = Scroller.get_scroller(w)



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui
from prettyqt.utils import datatypes, get_repr


if TYPE_CHECKING:
    import re

    from prettyqt.qt import QtCore


class BaseRegexValidator(gui.Validator):
    ID = "regex"

    def __init__(
        self,
        parent: QtCore.QObject | None = None,
        regex: datatypes.PatternType | None = None,
    ):
        super().__init__(parent)
        self.regex: re.Pattern | None = None
        if regex:
            self.set_regex(regex)

    def __repr__(self):
        return get_repr(self, self.get_regex())

    def __reduce__(self):
        return type(self), (self.get_regex(),)

    def __eq__(self, other: object):
        return self.regex == other.regex if isinstance(other, type(self)) else False

    def set_regex(self, regex: datatypes.PatternAndStringType):
        self.regex = datatypes.to_py_pattern(regex)

    def get_regex(self) -> str:
        if self.regex is None:
            msg = "Validator not initialized"
            raise TypeError(msg)
        return self.regex.pattern

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        if self.regex is None:
            msg = "Validator not initialized"
            raise TypeError(msg)
        if not text:
            return self.State.Intermediate, text, pos
        match = self.regex.match(text)  # type: ignore
        if match is None:
            #     return self.State.Invalid, text, pos
            # elif match.partial:  # type: ignore
            return self.State.Intermediate, text, pos
        return self.State.Acceptable, text, pos


class IntListValidator(BaseRegexValidator):
    """Validator which checks whether given string is a comma-separated list of ints."""

    ID = "int_list"

    def __init__(self, allow_single: bool = True, parent: QtCore.QObject | None = None):
        super().__init__(parent=parent)
        self.allow_single = allow_single
        if allow_single:
            self.set_regex(r"^(\d+)(,\s*\d+)*$")
        else:
            self.set_regex(r"^[0-9][0-9\,]+[0-9]$")

    def __reduce__(self):
        return type(self), (self.allow_single,)

    def __repr__(self):
        return get_repr(self, allow_single=self.allow_single)


class FloatListValidator(BaseRegexValidator):
    """Validator which checks whether given string is a comma-separated list of floats."""

    ID = "float_list"

    def __init__(self, allow_single: bool = True, parent: QtCore.QObject | None = None):
        super().__init__(parent=parent)
        self.allow_single = allow_single
        if allow_single:
            self.set_regex(r"^(\s*-?\d+(\.\d+)?)(\s*,\s*-?\d+(\.\d+)?)*$")
        else:
            self.set_regex(
                r"^(\s*-?\d+(\.\d+)?)(\s*,\s*-?\d+(\.\d+)?)(\s*,\s*-?\d+(\.\d+)?)*$"
            )

    def __reduce__(self):
        return type(self), (self.allow_single,)

    def __repr__(self):
        return get_repr(self, allow_single=self.allow_single)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.LineEdit()
    val = BaseRegexValidator()
    val.set_regex(r"\w\d\d")
    w.set_validator(val)
    w.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, gui


class SyntaxHighlighterMixin(core.ObjectMixin):
    def __init__(self, parent: core.QObject | None = None):
        super().__init__(parent)  # type: ignore

    def get_current_block(self) -> gui.TextBlock:
        return gui.TextBlock(self.currentBlock())

    def get_format(self, position: int) -> gui.TextCharFormat:
        return gui.TextCharFormat(self.format(position))


class SyntaxHighlighter(SyntaxHighlighterMixin, gui.QSyntaxHighlighter):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr


BoundaryReasonStr = Literal[
    "not_at_boundary",
    "break_opportunity",
    "start_of_item",
    "end_of_item",
    "mandatory_break",
    "soft_hyphen",
]

BOUNDARY_REASONS: bidict[BoundaryReasonStr, QtCore.QTextBoundaryFinder.BoundaryReason] = (
    bidict(
        not_at_boundary=QtCore.QTextBoundaryFinder.BoundaryReason.NotAtBoundary,
        break_opportunity=QtCore.QTextBoundaryFinder.BoundaryReason.BreakOpportunity,
        start_of_item=QtCore.QTextBoundaryFinder.BoundaryReason.StartOfItem,
        end_of_item=QtCore.QTextBoundaryFinder.BoundaryReason.EndOfItem,
        mandatory_break=QtCore.QTextBoundaryFinder.BoundaryReason.MandatoryBreak,
        soft_hyphen=QtCore.QTextBoundaryFinder.BoundaryReason.SoftHyphen,
    )
)

BoundaryTypeStr = Literal[
    "grapheme",
    "word",
    "line",
    "sentence",
]

BOUNDARY_TYPES: bidict[BoundaryTypeStr, QtCore.QTextBoundaryFinder.BoundaryType] = bidict(
    grapheme=QtCore.QTextBoundaryFinder.BoundaryType.Grapheme,
    word=QtCore.QTextBoundaryFinder.BoundaryType.Word,
    line=QtCore.QTextBoundaryFinder.BoundaryType.Line,
    sentence=QtCore.QTextBoundaryFinder.BoundaryType.Sentence,
)


class TextBoundaryFinder(QtCore.QTextBoundaryFinder):
    """Way of finding Unicode text boundaries in a string."""

    def __init__(
        self,
        string_or_other: str | QtCore.QTextBoundaryFinder = "",
        boundary_type: (
            QtCore.QTextBoundaryFinder.BoundaryType | BoundaryTypeStr
        ) = "grapheme",
    ):
        if isinstance(string_or_other, QtCore.QTextBoundaryFinder):
            super().__init__(string_or_other)
        else:
            if isinstance(boundary_type, str):
                typ = BOUNDARY_TYPES[boundary_type]
            else:
                typ = boundary_type
            super().__init__(typ, string_or_other)

    def __repr__(self):
        return get_repr(self, self.string())

    def __reduce__(self):
        return type(self), (self.string(), self.type())

    def __iter__(self):
        pos = self.position()
        self.setPosition(0)
        p = 0
        # if self.isAtBoundary():
        #     yield 0
        while p != -1:
            p = self.toNextBoundary()
            if p != -1:
                yield p
        self.setPosition(pos)

    def get_boundary_type(self) -> BoundaryTypeStr:
        return BOUNDARY_TYPES.inverse[self.type()]

    def get_boundary_reasons(self) -> list[BoundaryReasonStr]:
        return BOUNDARY_REASONS.get_list(self.boundaryReasons())


if __name__ == "__main__":
    reg = TextBoundaryFinder("  This is a test", boundary_type="sentence")
    for p in reg:
        print(p)



from __future__ import annotations

import pathlib
from typing import Literal

from prettyqt import core, gui


ModeStr = Literal["any", "file", "folder"]


class PathValidator(gui.Validator):
    """Validator which checks whether given string is a valid path."""

    ID = "path"

    def __init__(self, *args, **kwargs):
        self._mode = "any"
        super().__init__(*args, **kwargs)

    def __eq__(self, other: object):
        return isinstance(other, PathValidator) and self._mode == other._mode

    def validate(self, text: str, pos: int = 0):
        path = pathlib.Path(text)
        if not path.exists():
            return self.State.Intermediate, text, pos
        match self._mode:
            case "any":
                return self.State.Acceptable, text, pos
            case "file" if path.is_file():
                return self.State.Acceptable, text, pos
            case "folder" if path.is_dir():
                return self.State.Acceptable, text, pos
        return self.State.Intermediate, text, pos

    def set_mode(self, mode: ModeStr):
        self._mode = mode

    def get_mode(self) -> ModeStr:
        return self._mode

    mode = core.Property(
        str,
        get_mode,
        set_mode,
        doc="Whether text can be any, an existing file or an existing folder",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    val = PathValidator()
    app = widgets.app()
    widget = widgets.LineEdit("This is a test")
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, widgets


class CustomTitleBar(widgets.Widget):
    def __init__(self, parent, **kwargs):
        super().__init__(parent, **kwargs)
        layout = self.set_layout("horizontal", margin=0, spacing=0)
        layout.set_alignment("right")
        maximize_button = widgets.PushButton(clicked=parent.maximize)
        maximize_button.set_style_icon("titlebar_max_button", size=12)
        close_button = widgets.PushButton(clicked=parent.close)
        close_button.set_style_icon("titlebar_close_button", size=12)
        layout.add(widgets.Label(parent.windowTitle()))
        layout.add(maximize_button)
        layout.add(close_button)


class DockWidget(widgets.WidgetMixin, widgets.QDockWidget):
    """Widget that can be docked inside a QMainWindow ."""

    def __init__(self, *args, allowed_areas="all", **kwargs):
        super().__init__(*args, allowed_areas=allowed_areas, **kwargs)
        self._area: constants.DockWidgetAreasStr = "none"
        self.dockLocationChanged.connect(self._on_location_change)

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"allowedAreas": constants.DOCK_WIDGET_AREAS}
        return maps

    def set_widget(self, widget: widgets.QWidget):
        self.setWidget(widget)

    def set_allowed_areas(self, area: constants.DockWidgetAreasStr):
        self.setAllowedAreas(constants.DOCK_WIDGET_AREAS[area])

    def _on_location_change(self, area: constants.DockWidgetArea):
        self._area = constants.DOCK_WIDGET_AREAS.inverse[area]

    def get_current_area(self) -> constants.DockWidgetAreasStr:
        return self._area

    def setup_title_bar(self):
        title_bar = CustomTitleBar(parent=self)
        self.setTitleBarWidget(title_bar)

    def maximize(self):
        if not self.isFloating():
            self.setFloating(True)
        if not self.isMaximized():
            self.showMaximized()
        else:
            self.showMinimized()


if __name__ == "__main__":
    app = widgets.app()
    win = widgets.MainWindow()
    dock_widget = DockWidget(window_title="Test")
    dock_widget.setup_title_bar()
    win.add_dockwidget(dock_widget, "left")
    win.show()
    app.exec()



from __future__ import annotations

# from collections.abc import Sequence
import contextlib
import enum
from importlib import metadata
import logging
import pkgutil
from typing import ClassVar, Self

from packaging.markers import Marker
from packaging.requirements import InvalidRequirement, Requirement

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


def load_dist_info(name: str) -> metadata.Distribution | None:
    try:
        return metadata.distribution(name)
    except metadata.PackageNotFoundError:
        return None


def list_system_modules() -> list[metadata.Distribution]:
    modules = (test.name for test in pkgutil.iter_modules())
    distributions = (load_dist_info(i) for i in modules)
    return [i for i in distributions if i is not None]


def list_package_requirements(package_name: str) -> list[metadata.Distribution]:
    dist = metadata.distribution(package_name)
    try:
        modules = {Requirement(i).name for i in dist.requires} if dist.requires else set()
    except ValueError:
        logger.exception("Error listing requirements for %s", dist.name)
        return []
    return [dist for i in modules if (dist := load_dist_info(i)) is not None]


class DistTreeItem(itemmodels.ColumnItemModel.TreeItem):
    # __slots__ = ("requires", "metadata", "version", "markers")

    def __init__(
        self,
        obj,
        parent: DistTreeItem | None = None,
    ):
        super().__init__(obj, parent=parent)
        self.requires = {}
        self.markers = {}
        if obj is not None and obj.requires is not None:
            for i in obj.requires:
                with contextlib.suppress(InvalidRequirement):
                    req = Requirement(i)
                    self.requires[req.name] = req
                    if ";" in i:
                        extras_str = i.split(";", maxsplit=1)[-1]
                        self.markers[req.name] = Marker(extras_str)
        # Cache the data to avoid excessive IO (obj.metadata reads file)
        self.metadata = None if obj is None else obj.metadata
        self.version = None if obj is None else obj.version


class DistributionColumn(itemmodels.ColumnItem):
    def get_data(self, item: ImportlibTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case ImportlibTreeModel.Roles.DistributionRole:
                return item.obj


class NameColumn(DistributionColumn):
    name = "Name"
    doc = "Package name"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Name"] or ""
            case _:
                return super().get_data(item, role)


class VersionColumn(DistributionColumn):
    name = "Version"
    doc = "Version number."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.version or ""
            case _:
                return super().get_data(item, role)


class ConstraintsColumn(DistributionColumn):
    name = "Constraints"
    doc = "Constraints."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return next(
                    (
                        str(requirement.specifier)
                        for requirement in item.parent_item.requires.values()
                        if item.metadata["name"] == requirement.name
                    ),
                    "",
                )
            case _:
                return super().get_data(item, role)


class MarkerColumn(DistributionColumn):
    name = "Extra"
    doc = "Extra."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return next(
                    (
                        str(marker)
                        for name, marker in item.parent_item.markers.items()
                        if item.metadata["name"] == name
                    ),
                    "",
                )
            case _:
                return super().get_data(item, role)


class SummaryColumn(DistributionColumn):
    name = "Summary"
    doc = "Module description."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Summary"] or ""
            case _:
                return super().get_data(item, role)


class HomepageColumn(DistributionColumn):
    name = "Homepage"
    doc = "URL of the homepage."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Home-Page"] or ""
            case _:
                return super().get_data(item, role)


class AuthorColumn(DistributionColumn):
    name = "Author"
    doc = "Author name."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["Author"] or ""
            case _:
                return super().get_data(item, role)


class LicenseColumn(DistributionColumn):
    name = "License"
    doc = "License name."

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.metadata["License"] or ""
            case _:
                return super().get_data(item, role)


class ImportlibTreeModel(itemmodels.ColumnItemModel):
    """Model showing the dependency tree of a distribution."""

    @core.Enum
    class Roles(enum.IntEnum):
        DistributionRole = constants.USER_ROLE + 43255

    SUPPORTS = metadata.Distribution | str
    TreeItem = DistTreeItem
    IS_RECURSIVE = True
    COLUMNS: ClassVar = [
        NameColumn,
        VersionColumn,
        ConstraintsColumn,
        MarkerColumn,
        SummaryColumn,
        HomepageColumn,
        AuthorColumn,
        LicenseColumn,
    ]

    def __init__(
        self,
        obj: metadata.Distribution | str,
        show_root: bool = False,
        parent: core.QObject | None = None,
    ):
        if isinstance(obj, str):
            obj = metadata.distribution(obj)
        super().__init__(
            obj=obj, columns=self.COLUMNS, parent=parent, show_root=show_root
        )

    @classmethod
    def setup_example(cls):
        from prettyqt import widgets

        model = cls("prettyqt")
        table = widgets.TreeView(word_wrap=False)
        table.set_model(model)
        table.set_delegate("render_link", column=5)
        table.expand_all(depth=4)
        return table

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, metadata.Distribution)

    @classmethod
    def from_system(cls, parent: core.QObject | None = None) -> Self:
        distributions = list_system_modules()
        return cls(distributions, parent)

    @classmethod
    def from_package(cls, package_name: str, parent: core.QObject | None = None) -> Self:
        distributions = list_package_requirements(package_name)
        return cls(distributions, parent)

    def _has_children(self, item: ImportlibTreeModel.TreeItem) -> bool:
        return bool(item.requires)

    def _fetch_object_children(self, item: DistTreeItem) -> list[DistTreeItem]:
        return [
            DistTreeItem(obj=dist, parent=item)
            for dist in list_package_requirements(item.metadata["Name"])
        ]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    model = ImportlibTreeModel("prettyqt")
    table = widgets.TreeView(word_wrap=False)
    # table.setSortingEnabled(True)
    table.set_model(model)
    table.set_delegate("render_link", column=5)
    table.expand_all(depth=4)
    table.show()
    app.exec()



from __future__ import annotations

from prettyqt import scxml


class ScxmlCppDataModel(
    scxml.scxmldatamodel.ScxmlDataModelMixin, scxml.QScxmlCppDataModel
):
    """C++ data model for a Qt SCXML state machine."""



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


CacheModeStr = Literal["none", "all"]

CACHE_MODE: bidict[CacheModeStr, QtGui.QMovie.CacheMode] = bidict(
    none=QtGui.QMovie.CacheMode.CacheNone, all=QtGui.QMovie.CacheMode.CacheAll
)

MovieStateStr = Literal["not_running", "paused", "running"]

MOVIE_STATE: bidict[MovieStateStr, QtGui.QMovie.MovieState] = bidict(
    not_running=QtGui.QMovie.MovieState.NotRunning,
    paused=QtGui.QMovie.MovieState.Paused,
    running=QtGui.QMovie.MovieState.Running,
)


class Movie(core.ObjectMixin, QtGui.QMovie):
    """Convenience class for playing movies with QImageReader."""

    def __repr__(self):
        return get_repr(self, self.fileName(), self.get_format())

    def set_cache_mode(self, mode: CacheModeStr | QtGui.QMovie.CacheMode):
        """Set cache mode.

        Args:
            mode: cache mode
        """
        self.setCacheMode(CACHE_MODE.get_enum_value(mode))

    def get_cache_mode(self) -> CacheModeStr:
        """Get the current cache mode.

        Returns:
            cache mode
        """
        return CACHE_MODE.inverse[self.cacheMode()]

    def get_state(self) -> MovieStateStr:
        """Get the current state.

        Returns:
            state
        """
        return MOVIE_STATE.inverse[self.state()]

    def get_format(self) -> str:
        return self.format().data().decode()

    @classmethod
    def get_supported_formats(cls) -> list[str]:
        return [i.data().decode() for i in cls.supportedFormats()]


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    movie = Movie()
    print(repr(movie.get_format()))



from __future__ import annotations

import html
import os
from typing import TYPE_CHECKING, Any, Self

from prettyqt import constants, core, gui, iconprovider
from prettyqt.utils import datatypes, get_repr, helpers, listdelegators, serializemixin


if TYPE_CHECKING:
    from collections.abc import Iterator


class StandardItem(serializemixin.SerializeMixin, gui.QStandardItem):
    def __repr__(self):
        return get_repr(self, self.get_icon(), self.text())

    def __getitem__(
        self, index: int | slice | tuple[int | slice, int | slice] | core.QModelIndex
    ) -> gui.QStandardItem | listdelegators.ListDelegator[gui.QStandardItem]:
        match index:
            case int():
                if index >= self.childCount():
                    raise IndexError(index)
                return self.child(index)
            case slice():
                return self.__getitem__(index, 0)
            case int() as row, int() as col:
                return self.child(row, col)
            case (row, col):
                rowcount = self.rowCount()
                colcount = self.columnCount()
                children = [
                    self.child(i, j)
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                ]
                return listdelegators.ListDelegator(children)
            case _:
                raise TypeError(index)

    def __delitem__(self, index: int | tuple[int, int]):
        item = self.takeRow(index) if isinstance(index, int) else self.takeChild(*index)
        if item is None:
            raise KeyError(index)
        return item

    def __iter__(self) -> Iterator[gui.QStandardItem]:
        return iter(self.get_children())

    def __add__(self, other: str | gui.QStandardItem) -> StandardItem:
        match other:
            case gui.QStandardItem() | str():
                self.add(other)
                return self
            case _:
                msg = "wrong type for addition"
                raise TypeError(msg)

    def get_children(self) -> listdelegators.ListDelegator[gui.QStandardItem]:
        items = [self.child(index) for index in range(self.rowCount())]
        return listdelegators.ListDelegator(items)

    def add(self, *item: str | gui.QStandardItem):
        for i in item:
            new_item = type(self)(i) if isinstance(i, str) else i
            self.appendRow([new_item])

    def clone(self) -> Self:
        item = type(self)()
        core.DataStream.copy_data(self, item)
        return item

    def set_icon(self, icon: datatypes.IconType):
        """Set the icon for the action.

        Args:
            icon: icon to use
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(icon)

    def set_checkstate(
        self, state: constants.CheckStateStr | constants.CheckState | bool
    ):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use
        """
        if isinstance(state, bool):
            state = (
                constants.CheckState.Checked if state else constants.CheckState.Unchecked
            )
        self.setCheckState(constants.CHECK_STATE.get_enum_value(state))

    def is_checked(self):
        return self.checkState() == constants.CheckState.Checked

    def toggle_checkstate(self):
        if self.checkState() == constants.CheckState.Checked:
            self.setCheckState(constants.CheckState.Unchecked)
        else:
            self.setCheckState(constants.CheckState.Checked)

    def get_checkstate(self) -> constants.CheckStateStr:
        """Return checkstate.

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState()]

    def set_text_alignment(
        self, alignment: constants.AlignmentStr | constants.AlignmentFlag
    ):
        """Set the alignment of the text.

        Args:
            alignment: alignment for the format
        """
        self.setTextAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_text_alignment(self) -> constants.AlignmentStr:
        """Return current text alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.textAlignment()]

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def get_foreground(self) -> gui.Brush:
        return gui.Brush(self.foreground())

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())

    def get_icon(self) -> gui.Icon | None:
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_data(self, data: Any, role: constants.ItemDataRoleStr | int):
        item_role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        super().setData(data, item_role)

    def get_data(self, role: constants.ItemDataRoleStr | int):
        item_role = constants.ITEM_DATA_ROLE[role] if isinstance(role, str) else role
        return super().data(item_role)

    def set_tooltip(
        self,
        tooltip: str | datatypes.PathType,
        size: datatypes.SizeType | None = None,
        rich_text: bool = False,
    ):
        if isinstance(tooltip, os.PathLike):
            path = os.fspath(tooltip)
            if size is None:
                tooltip = f"<img src={path!r}>"
            else:
                s = datatypes.to_size(size)
                tooltip = f'<img src={path!r} width="{s.width()}" height="{s.height()}">'
        if rich_text:
            tooltip = f"<html>{html.escape(tooltip)}</html>"
        super().setToolTip(tooltip)

    def set_size_hint(self, hint: datatypes.SizeType):
        self.setSizeHint(datatypes.to_size(hint))

    def add_item(
        self,
        name: str = "",
        icon: datatypes.IconType = None,
        data: dict | None = None,
        foreground: gui.QBrush | None = None,
        background: gui.QBrush | None = None,
        font: gui.QFont | None = None,
        selectable: bool = True,
        enabled: bool = True,
        editable: bool = False,
        status_tip: str = "",
        tool_tip: str = "",
        whats_this: str | None = None,
        # text_alignment: Optional[str] = None,
        checkstate: constants.CheckStateStr | None = None,
        flags: constants.ItemFlag | None = None,
        size_hint: datatypes.SizeType | None = None,
        is_user_type: bool = False,
    ) -> Self:
        item = type(self)(name)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.set_data(v, k)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if enabled:
            item.setEnabled(enabled)
        if editable:
            item.setEditable(editable)
        if selectable:
            item.setSelectable(selectable)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.set_size_hint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.appendRow([item])
        return item


if __name__ == "__main__":
    item = StandardItem()
    item.set_data("test", 1000)
    item2 = StandardItem()
    item.add(item2)
    print(item.child(0))



from __future__ import annotations

from typing import Literal

from prettyqt import core, widgets
from prettyqt.utils import bidict


CHANGE_FLAGS = bidict(
    scale_factor=widgets.QPinchGesture.ChangeFlag.ScaleFactorChanged,
    rotation_angle=widgets.QPinchGesture.ChangeFlag.RotationAngleChanged,
    center_point=widgets.QPinchGesture.ChangeFlag.CenterPointChanged,
)

ChangeFlagStr = Literal["scale_factor", "rotation_angle", "center_point"]


class PinchGesture(widgets.GestureMixin, widgets.QPinchGesture):
    """Describes a pinch gesture made by the user."""

    def get_start_center_point(self) -> core.PointF:
        return core.PointF(self.startCenterPoint())

    def get_center_point(self) -> core.PointF:
        return core.PointF(self.centerPoint())

    def get_last_center_point(self) -> core.PointF:
        return core.PointF(self.lastCenterPoint())

    def get_change_flags(self) -> list[ChangeFlagStr]:
        return CHANGE_FLAGS.get_list(self.changeFlags())

    def set_change_flags(self, **kwargs):
        val = widgets.QPinchGesture.ChangeFlag(0)
        for k, v in kwargs.items():
            if v is True:
                val |= CHANGE_FLAGS[k]
        flag = widgets.QPinchGesture.ChangeFlag(val)  # type: ignore
        self.setChangeFlags(flag)  # type: ignore

    def get_total_change_flags(self) -> list[ChangeFlagStr]:
        return CHANGE_FLAGS.get_list(self.totalChangeFlags())

    def set_total_change_flags(self, **kwargs):
        val = widgets.QPinchGesture.ChangeFlag(0)
        for k, v in kwargs.items():
            if v is True:
                val |= CHANGE_FLAGS[k]
        flag = widgets.QPinchGesture.ChangeFlag(val)  # type: ignore
        self.setTotalChangeFlags(flag)  # type: ignore


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    gesture = PinchGesture()



"""Tests for `prettyqt` package."""

import logging
import re

import pytest

import prettyqt
from prettyqt import constants, core, custom_widgets, gui, widgets
import prettyqt.custom_widgets.dataset as fo


def test_booldicttoolbutton(qtbot):
    dct = dict(a="test", b="test2")
    w = custom_widgets.BoolDictToolButton(text="Title", dct=dct)
    qtbot.addWidget(w)
    w["a"] = True
    assert w["a"] is True
    assert w.as_dict() == dict(a=True, b=False)


def test_collapsibleframe(qtbot):
    frame = custom_widgets.CollapsibleFrame()
    qtbot.addWidget(frame)
    widget = widgets.Widget()
    qtbot.addWidget(widget)
    frame.add_widget(widget)
    frame.set_title("CollapsibleFrame")
    frame.expand()
    assert frame.is_expanded() is True
    frame.collapse()
    frame.remove_widget(widget)


def test_elidedlabel(qtbot):
    label = custom_widgets.ElidedLabel()
    label.set_text("test")
    assert label.get_text() == "test"
    label.show()


# def test_expandableline(qtbot):
#     layout = widgets.VBoxLayout()
#     textbrowser = widgets.TextBrowser()
#     # qtbot.addWidget(textbrowser)
#     layout.addWidget(textbrowser)
#     widget = custom_widgets.ExpandableLine("Test")
#     qtbot.addWidget(widget)
#     widget.set_layout(layout)
#     widget.show()


def test_colorchooserbutton(qtbot):
    btn = custom_widgets.ColorChooserButton()
    qtbot.addWidget(btn)
    btn.set_current_color("green")
    assert btn.get_value() == gui.Color("green")
    btn.set_value("blue")
    assert btn.is_valid()


@pytest.mark.skipif(prettyqt.qt.API == "pyside6", reason="Segmentation fault")
def test_regexeditor(qtbot):
    widget = custom_widgets.regexeditor.RegexEditorWidget()
    qtbot.addWidget(widget)
    widget.regex = "[0-9]"
    assert widget.regex == "[0-9]"
    widget.string = "test123"
    assert widget.string == "test123"
    widget.compile_flags = re.IGNORECASE


def test_dataset(qtbot):
    class Test(fo.DataSet):
        i1 = fo.Bool(label="My first one")
        string1 = fo.String(label="My first one", regex="[0-9]")
        string2 = fo.String(label="My second one", notempty=True)
        choiceitem = fo.Enum(label="A", choices=["A", "B"], disabled_on="i1")
        mchoiceitem = fo.MultipleChoice(label="A", choices=["A", "B"])
        floatitem = fo.Float(label="FloatItem", enabled_on="i1")
        intitem = fo.Int(label="IntItem", enabled_on="i1")
        coloritem = fo.Color(label="ColorItem", value="green")
        fileitem = fo.File(label="File")
        folderitem = fo.Folder(label="File", optional="Test")
        stringornumber = fo.StringOrNumber(label="Test")
        buttonitem = fo.Button(label="Button", callback=print)
        intlist = fo.IntList(label="Button", value=[1, 2, 3])
        floatlist = fo.FloatList(label="Button", value=[1.1, 2, 3])
        code = fo.Code(label="Test", value="class Test")

    settings = Test(icon="mdi.timer")
    settings.create_dialog()
    settings.to_dict()


def test_flagselectionwidget(qtbot):
    widget = custom_widgets.FlagSelectionWidget()
    qtbot.addWidget(widget)
    items = {0: "MultiLine", 2: "Ignore case"}
    widget.add_items(items)
    assert widget.get_value() == 0


def test_iconbrowser(qtbot):
    browser = custom_widgets.IconBrowser()
    qtbot.addWidget(browser)
    browser.show()
    browser.hide()


def test_stringornumberwidget(qtbot):
    widget = custom_widgets.StringOrNumberWidget()
    qtbot.addWidget(widget)
    widget.get_value()
    widget.on_value_change()


def test_optionalwidget(qtbot):
    w = widgets.RadioButton()
    qtbot.addWidget(w)
    container = custom_widgets.OptionalWidget(w, "Test")
    qtbot.addWidget(container)
    container.get_value()
    container.enabled = False
    assert container.enabled is False


def test_sidebarwidget(qtbot):
    ex = custom_widgets.SidebarWidget(show_settings=True)
    qtbot.addWidget(ex)
    page_1 = widgets.PlainTextEdit()
    page_2 = widgets.ColorDialog()
    page_3 = widgets.FileDialog()
    qtbot.addWidget(page_1)
    qtbot.addWidget(page_2)
    qtbot.addWidget(page_3)
    ex.add_tab(page_1, "Text", "mdi.timer")
    ex.add_tab(page_2, "Color", "mdi.format-color-fill", area="bottom")
    ex.add_tab(page_3, "Help", "mdi.help-circle-outline")
    ex.set_marker(page_3)
    ex.set_tab(page_2)


def test_singlelinetextedit(qtbot):
    w = custom_widgets.SingleLineTextEdit()
    qtbot.addWidget(w)
    w.set_text("test")


def test_timeline(qtbot):
    tl = custom_widgets.Timeline(60, 60)
    qtbot.addWidget(tl)
    tl.show()
    sample = custom_widgets.VideoSample(20)
    tl += sample
    tl.add_sample(30)
    sample_3 = custom_widgets.VideoSample(20)
    tl[1] = sample_3
    assert tl[1] == sample_3
    assert len(tl) == 2
    tl.close()


def test_mappedcheckbox(qtbot):
    widget = custom_widgets.MappedCheckBox(true_value=0, false_value=1)
    qtbot.addWidget(widget)
    widget.set_value(0)
    assert widget.get_value() == 0
    widget.setChecked(False)
    assert widget.get_value() == 1


def test_filechooserbutton(qtbot):
    btn = custom_widgets.FileChooserButton()
    qtbot.addWidget(btn)
    btn.set_value("/")
    btn.get_value()


# Segfault with PySide6
# def test_objectbrowser(qtbot):
#     struct = dict(a={1, 2, frozenset([1, 2])})
#     browser = custom_widgets.ObjectBrowser(struct)
#     browser.show()
#     browser.close()


# Segfault with PySide6
# def test_fontchooserbutton(qtbot):
#     btn = custom_widgets.FontChooserButton()
#     qtbot.addWidget(btn)
#     btn.set_font("Consolas")


def test_iconlabel(qtbot):
    w = custom_widgets.IconLabel()
    qtbot.addWidget(w)
    w.set_text("test")
    assert w.text() == "test"
    repr(w)


def test_inputandslider(qtbot):
    w = custom_widgets.InputAndSlider()
    qtbot.addWidget(w)
    w.set_range(0, 10)
    w.set_value(5)
    assert w.get_value() == 5
    w.set_step_size(2)
    assert w.is_valid()


# something related to syntaxhighlighting...
@pytest.mark.skipif(
    prettyqt.qt.API == "pyside6", reason="Windows fatal exception: access violation"
)
def test_codeeditor(qtbot):
    editor = custom_widgets.CodeEditor()
    qtbot.addWidget(editor)
    assert not editor.text()
    editor.line_area_width()
    editor.set_syntaxhighlighter("python")
    event = gui.QResizeEvent(core.Size(10, 10), core.Size(20, 20))
    editor.resizeEvent(event)
    editor.repaint()


def test_flowlayout(qtbot):
    widget = widgets.Widget()
    qtbot.addWidget(widget)
    layout = custom_widgets.FlowLayout(margin=1)
    btn = widgets.PushButton("Short")
    qtbot.addWidget(btn)
    layout += btn
    layout += widgets.PushButton("Longer")
    layout += widgets.PushButton("Different text")
    layout += widgets.PushButton("More text")
    layout += widgets.PushButton("Even longer button text")
    layout._do_layout(core.Rect(), False)
    layout.sizeHint()
    widget.set_layout(layout)
    assert layout[0] == btn
    for _i in layout:
        pass
    assert len(layout) == 5
    layout.get_children()


def test_labeledslider(qtbot, qttester):
    slider = custom_widgets.LabeledSlider(["test1", "test2"], "vertical")
    slider_2 = custom_widgets.LabeledSlider(["test1", "test2"])
    qtbot.addWidget(slider)
    qtbot.addWidget(slider_2)
    slider.show()
    qtbot.add_widget(slider)
    qttester.send_mousepress(slider.sl, constants.MouseButton.LeftButton)
    qttester.send_mousemove(slider.sl, core.PointF(20, 20))
    slider.repaint()
    slider.hide()


def test_logtextedit(qtbot):
    textedit = custom_widgets.LogTextEdit()
    qtbot.addWidget(textedit)
    textedit.show()
    logger = logging.getLogger()
    fmt = logging.Formatter(
        "%(asctime)s  %(levelname)s %(levelno)i  %(message)s %(filename)s "
        "%(funcName)s %(module)s %(created)f %(lineno)d %(msecs)d "
        "%(process)d %(thread)d %(threadName)s %(processName)s "
        "%(relativeCreated)s %(name)s %(pathname)s"
    )
    textedit.set_formatter(fmt)
    logger.info("Test")
    try:
        raise Exception
    except Exception as e:
        logger.exception(e)
    textedit.hide()


def test_popupinfo(qtbot):
    popup = custom_widgets.PopupInfo()
    qtbot.addWidget(popup)
    popup.show_popup("test")


def test_roundprogressbar(qtbot):
    bar = custom_widgets.RoundProgressBar()
    qtbot.addWidget(bar)
    bar.show()
    assert bar.minimum() == 0
    assert bar.maximum() == 100
    bar.set_value(50)
    assert bar.get_value() == 50
    bar.set_null_position(40)
    bar.hide()
    bar.set_bar_style("donut")
    bar.set_outline_pen_width(3)
    bar.set_data_pen_width(5)
    bar.set_format(r"%p")
    bar.set_decimals(3)
    bar.set_range(20, 80)


def test_selectionwidget(qtbot):
    widget = custom_widgets.SelectionWidget()
    qtbot.addWidget(widget)

    class Test:
        pass

    test = Test()
    items = {";": "Semicolon", "tab": "Tab", ",": "Comma", test: "class"}
    widget.add_items(items)
    widget.add_items(("a", "b"))
    widget.add_tooltip_icon("test")
    widget.add_custom(label="test", regex=r"\S{1}")
    radiobuttons = [k for k, v in widget.buttons.items()]
    radiobuttons[1].click()
    assert widget.get_value() == "tab"
    radiobuttons[3].click()
    assert widget.current_choice() == test
    for _i in widget:
        pass
    widget.select_radio_by_data(";")
    choice = widget.current_choice()
    assert choice == ";"
    widget.set_value(",")
    assert widget.get_value() == ","
    widget.update_choice(True)


def test_spanslider(qtbot, qttester):
    slider = custom_widgets.SpanSlider()
    qtbot.add_widget(slider)
    slider.show()
    slider.set_lower_value(10)
    slider.set_upper_value(20)
    slider.set_lower_pos(15)
    slider.set_lower_pos(15)
    slider.set_upper_pos(25)
    slider.set_upper_pos(25)
    assert slider.lower_value == 15
    assert slider.upper_value == 25
    slider.set_value((16, 24))
    assert slider.get_value() == (16, 24)
    slider.set_lower_value(12)
    slider.set_upper_pos(20)
    color = gui.Color("blue")
    slider.set_left_color(color)
    slider.set_right_color(color)
    slider._swap_controls()
    slider.trigger_action(slider.SliderAction.SliderNoAction, True)
    slider.trigger_action(slider.SliderAction.SliderSingleStepAdd, True)
    slider.repaint()
    slider._pixel_pos_to_value(100)
    slider._move_pressed_handle()
    qttester.send_mousepress(slider, constants.MouseButton.LeftButton)
    qttester.send_mousemove(slider, core.PointF(20, 20))
    qttester.send_mousemove(slider, core.PointF(0, 0), delay=10)
    assert slider.get_movement_mode() == "no_crossing"
    slider.set_movement_mode("no_overlap")
    slider.close()


def test_waitingspinner(qtbot):
    test_widget = widgets.Widget()
    qtbot.addWidget(test_widget)
    spinner = custom_widgets.WaitingSpinner(parent=test_widget)
    qtbot.addWidget(spinner)
    spinner.repaint()
    spinner.set_line_num(2)
    assert spinner.get_line_num() == 2
    spinner.set_line_length(2)
    assert spinner.get_line_length() == 2
    spinner.set_line_width(2)
    assert spinner.get_line_width() == 2
    spinner.set_inner_radius(2)
    assert spinner.get_inner_radius() == 2
    spinner.set_color("black")
    spinner.set_revolutions_per_second(2)
    spinner.set_trail_fade_percentage(2)
    spinner.set_minimum_trail_opacity(2)
    spinner._rotate()
    spinner.start()
    spinner.stop()
    spinner._update_position()



from __future__ import annotations

from .baseeventfilter import BaseEventFilter
from .animatedtooltipeventfilter import AnimatedToolTipEventFilter
from .autosizecolumnseventfilter import AutoSizeColumnsEventFilter
from .eventcatcher import EventCatcher
from .hovericoneventfilter import HoverIconEventFilter
from .listviewgridresizeeventfilter import ListViewGridResizeEventFilter
from .sectionautospaneventfilter import SectionAutoSpanEventFilter
from .slidermovetomouseclickeventfilter import SliderMoveToMouseClickEventFilter
from .timelabeleventfilter import (
    DateLabelEventFilter,
    TimeLabelEventFilter,
    TextUpdateEventFilter,
)


__all__ = [
    "AnimatedToolTipEventFilter",
    "AutoSizeColumnsEventFilter",
    "BaseEventFilter",
    "DateLabelEventFilter",
    "EventCatcher",
    "HoverIconEventFilter",
    "ListViewGridResizeEventFilter",
    "SectionAutoSpanEventFilter",
    "SliderMoveToMouseClickEventFilter",
    "TextUpdateEventFilter",
    "TimeLabelEventFilter",
]



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


TYPES = bidict(
    variable=QtGui.QTextLength.Type.VariableLength,
    fixed=QtGui.QTextLength.Type.FixedLength,
    percentage=QtGui.QTextLength.Type.PercentageLength,
)

TypeStr = Literal["variable", "fixed", "percentage"]


class TextLength(QtGui.QTextLength):
    def __repr__(self):
        return get_repr(self, self.type(), self.rawValue())

    def get_type(self) -> TypeStr:
        """Return type of this length object.

        Returns:
            timer type
        """
        return TYPES.inverse[self.type()]


if __name__ == "__main__":
    length = TextLength()



from __future__ import annotations

import logging

from prettyqt import constants, widgets
from prettyqt.utils import helpers, listdelegators


logger = logging.getLogger(__name__)


class GridSplitter(widgets.Splitter):
    """Nested Splitter which keeps sections in sync.

    Currently only works properly if it is a proper grid.
    In future, this could be expanded to work with spans.
    """

    def __init__(self, **kwargs):
        self.sub_splitters = []
        super().__init__(constants.VERTICAL, **kwargs)

    def __getitem__(
        self, index: tuple[int | slice, int | slice] | int | str
    ) -> widgets.QWidget | listdelegators.ListDelegator[widgets.QWidget] | None:
        rowcount = self.rowCount()
        colcount = self.columnCount()
        match index:
            case int() as row, int() as col:
                if row >= rowcount or col >= colcount:
                    raise IndexError(index)
                return self.itemAtPosition(row, col)
            case (row, col):
                items = [
                    item
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                    if (item := self.itemAtPosition(i, j)) is not None
                ]
                return listdelegators.ListDelegator(list(set(items)))
            case int() as row:
                if row >= rowcount:
                    raise IndexError(index)
                return self.itemAt(row)
            case slice() as rowslice:
                count = rowcount if rowslice.stop is None else rowslice.stop
                items = [self.itemAt(i) for i in range(count)[rowslice]]
                return listdelegators.ListDelegator(list(set(items)))
            case str():
                return self.find_child(widgets.QWidget, index)
            case _:
                raise TypeError(index)

    def rowCount(self) -> int:
        return len(self.sub_splitters)

    def columnCount(self) -> int:
        return max(splitter.count() for splitter in self.sub_splitters)

    def itemAtPosition(self, row: int, col: int) -> widgets.QWidget:
        return self.sub_splitters[row].widget(col)

    def itemAt(self, row: int) -> widgets.QWidget:
        widgets = [w for s in self.sub_splitters for w in s.get_widgets()]
        return widgets[row]

    def __setitem__(
        self,
        idx: tuple[int, int],
        widget: widgets.QWidget,
    ):
        row, col = idx
        self.add(widget, row, col)

    def add(self, widget: widgets.QWidget, row: int, column: int):
        while len(self.sub_splitters) <= row:
            self._add_sub_splitter()
        splitter = self.sub_splitters[row]
        if len(splitter) > column:
            splitter.replaceWidget(column, widget)
        else:
            splitter.add(widget)
        # self.align_sizes()

    def _add_sub_splitter(self):
        splitter = widgets.Splitter(constants.HORIZONTAL)
        super().add(splitter)
        splitter.splitterMoved.connect(self._align_sizes)
        self.sub_splitters.append(splitter)

    def _on_handle_move(self, position: int, index: int):
        logger.debug(f"{position=} {index=}")
        sender = self.sender()
        receivers = [splitter for splitter in self.sub_splitters if splitter != sender]
        for rec in receivers:
            handle = rec.handle(index)
            if handle is not None:
                with rec.signals_blocked():
                    handle.moveSplitter(position)

    def show(self):
        super().show()
        # self.align_sizes()

    def _align_sizes(self):
        self.refresh()
        sizes = [splitter.sizes() for splitter in self.sub_splitters]
        max_sizes = [max(i) for i in zip(*sizes)]
        for splitter in self.sub_splitters:
            # for i, w in enumerate(splitter.get_widgets()):
            #     if i < len(max_sizes):
            #         w.resize(max_sizes[i], w.height())
            splitter.setSizes(max_sizes[: splitter.count()])


if __name__ == "__main__":
    app = widgets.app()
    splitter = GridSplitter()
    splitter.resize(500, 500)
    with app.debug_mode():
        splitter[0, 0] = widgets.Label("0, 0")
        splitter[0, 1] = widgets.Label("0, 1")
        # splitter[0, 2] = widgets.Label("0, 2")
        splitter[1, 0] = widgets.Label("1, 0")
        splitter[1, 1] = widgets.Label("1, 1")
        splitter[2, 0] = widgets.Label("2, 0")
        splitter[2, 1] = widgets.Label("2, 1")
        splitter.show()
        app.exec()



from __future__ import annotations

from prettyqt import constants, core, gui, svg
from prettyqt.utils import colors, datatypes


class SVGBufferIconEngine(gui.IconEngine):
    """A custom QIconEngine that can render an SVG buffer.

    An icon engine provides the rendering functions for a ``QIcon``.
    Each icon has a corresponding icon engine that is responsible for drawing
    the icon with a requested size, mode and state.  While the built-in
    QIconEngine is capable of rendering SVG files, it's not able to receive the
    raw XML string from memory.
    This ``QIconEngine`` takes in SVG data as a raw xml string or bytes.
    see: https://doc.qt.io/qt-5/qiconengine.html
    """

    def __init__(self, xml: str, color: datatypes.ColorType = "black") -> None:
        self._xml = xml
        self._color = colors.get_color(color)
        super().__init__()

    def paint(
        self,
        painter: gui.QPainter,
        rect: core.QRect,
        mode: gui.QIcon.Mode,
        state: gui.QIcon.State,
    ):
        """Paint the icon int ``rect`` using ``painter``."""
        color = self._color
        if mode == gui.QIcon.Mode.Disabled:
            color = self._color.transparent(0.3)

        xml = self._xml.replace('fill="currentColor"', color.get_name("svg_argb"))
        xml_byte = xml.encode()
        renderer = svg.SvgRenderer(xml_byte)
        renderer.render(painter, core.RectF(rect))

    def clone(self):
        """Required to subclass abstract QIconEngine."""
        return SVGBufferIconEngine(self._xml, self._color)

    def pixmap(
        self, size: core.QSize, mode: gui.QIcon.Mode, state: gui.QIcon.State
    ) -> gui.QPixmap:
        """Return the icon as a pixmap with requested size, mode, and state."""
        img = gui.Image(size, gui.QImage.Format.Format_ARGB32)
        img.fill(constants.GlobalColor.transparent)
        pixmap = gui.QPixmap.fromImage(
            img, constants.ImageConversionFlag.NoFormatConversion
        )
        rect = core.QRect(core.QPoint(0, 0), size)
        self.paint(gui.QPainter(pixmap), rect, mode, state)
        return pixmap

    def change_color(self, color: datatypes.ColorType) -> None:
        self._color = colors.get_color(color)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    engine = SVGBufferIconEngine("")
    widget = widgets.TreeView()
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Any

from prettyqt import constants, core


class ChangeHeadersProxyModel(core.IdentityProxyModel):
    """Proxy model for changing the header data (either horizontal or vertical).

    Header data can either be changed by passing a list with same length as source length
    or by passing a dictionary with index as key and new value as value
    (Example: {1: "abc", 3: "def"} changes section 1 to "abc" and section 3 to "def")
    Apart from the regular use case of changing the text, the other roles can be changed,
    too.

    ### Example

    ```py
    table.proxifier.change_headers(
        header=["x", "y", "z"],
        orientation=constants.HORIZONTAL,
        role=constants.DISPLAY_ROLE,
    )
    table.show()
    # or
    model = MyModel()
    proxy = ChangeHeadersProxyModel(
        header=["x", "y", "z"], orientation=constants.VERTICAL
    )
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "change_headers"
    ICON = "table-headers-eye"

    def __init__(
        self,
        header: list[Any] | dict[int, Any],
        orientation: constants.Orientation
        | constants.OrientationStr = constants.HORIZONTAL,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._orientation = constants.ORIENTATION.get_enum_value(orientation)
        self._header = header
        self._role = role

    def setSourceModel(self, model):
        header_len = (
            model.columnCount()
            if self._orientation == constants.HORIZONTAL
            else model.rowCount()
        )
        if isinstance(self._header, list) and len(self._header) != header_len:
            msg = "list needs to be same list as header"
            raise ValueError(msg)
        super().setSourceModel(model)

    def get_header(self) -> list[int]:
        return self._header

    def set_header(
        self,
        header: list[str] | dict[int, str],
        orientation: constants.Orientation | constants.OrientationStr,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        with self.reset_model():
            self._header = header
            self._orientation = constants.ORIENTATION.get_enum_value(orientation)
            self._role = role

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if orientation == self._orientation and role == self._role:  # noqa: SIM102
            if (isinstance(self._header, dict) and section in self._header) or isinstance(
                self._header, list
            ):
                return self._header[section]
        return self.sourceModel().headerData(section, orientation, role)

    header = core.Property(
        object,
        get_header,
        set_header,
        doc="New headers (dict / list)",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    data = dict(
        a=["abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedfaa"],
        b=[10000000, 2, 3, 4, 5, 6],
        c=[1, 2, 3, 4, 5, 6],
        d=[100000000, 2, 3, 4, 5, 6],
        e=[1000000, 2, 3, 4, 5, 6],
    )
    source_model = gui.StandardItemModel.from_dict(data)
    app = widgets.app()
    table = widgets.TableView()
    model = ChangeHeadersProxyModel(
        {0: constants.ALIGN_RIGHT}, role=constants.ALIGNMENT_ROLE, parent=table
    )
    model.setSourceModel(source_model)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.utils import bidict, datatypes


ImageReaderErrorStr = Literal[
    "file_not_found", "device", "unsupported_format", "invalid_data", "unknown"
]

IMAGE_READER_ERROR: bidict[ImageReaderErrorStr, gui.QImageReader.ImageReaderError] = (
    bidict(
        file_not_found=gui.QImageReader.ImageReaderError.FileNotFoundError,
        device=gui.QImageReader.ImageReaderError.DeviceError,
        unsupported_format=gui.QImageReader.ImageReaderError.UnsupportedFormatError,
        invalid_data=gui.QImageReader.ImageReaderError.InvalidDataError,
        unknown=gui.QImageReader.ImageReaderError.UnknownError,
    )
)


class ImageReader(gui.QImageReader):
    """Format independent interface for reading images from files or other devices."""

    def __getitem__(self, key: str) -> str:
        return self.text(key)

    def get_error(self) -> ImageReaderErrorStr:
        """Return error type.

        Returns:
            error type
        """
        return IMAGE_READER_ERROR.inverse[self.error()]

    def get_background_color(self) -> gui.Color:
        return gui.Color(self.backgroundColor())

    def get_clip_rect(self) -> core.Rect:
        return core.Rect(self.clipRect())

    def get_current_image_rect(self) -> core.Rect:
        return core.Rect(self.currentImageRect())

    def get_scaled_clip_rect(self) -> core.Rect:
        return core.Rect(self.scaledClipRect())

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def get_scaled_size(self) -> core.Size:
        return core.Size(self.scaledSize())

    def get_format(self) -> str:
        return self.format().data().decode()

    def get_subtype(self) -> str:
        return self.subType().data().decode()

    def get_supported_subtypes(self) -> list[str]:
        return [i.data().decode() for i in self.supportedSubTypes()]

    def set_format(self, fmt: datatypes.ByteArrayType):
        fmt = datatypes.to_bytearray(fmt)
        self.setFormat(fmt)

    def get_transformation(self) -> gui.imageiohandler.TransformationStr:
        """Return the transformation and orientation the image has been set to.

        Returns:
            transformation
        """
        return gui.imageiohandler.TRANSFORMATION.inverse[self.transformation()]

    def read_image(self) -> gui.Image:
        return gui.Image(self.read())

    def supports_option(
        self, option: gui.imageiohandler.ImageOptionStr | gui.QImageIOHandler.ImageOption
    ) -> bool:
        """Return whether the image handler supports given option.

        Args:
            option: option to check

        Returns:
            option
        """
        return self.supportsOption(gui.imageiohandler.IMAGE_OPTION.get_enum_value(option))

    @staticmethod
    def get_image_format(obj: str | core.QIODevice) -> str:
        return ImageReader.imageFormat(obj).data().decode()

    @staticmethod
    def get_supported_image_formats() -> list[str]:
        return [i.data().decode() for i in ImageReader.supportedImageFormats()]

    @staticmethod
    def get_supported_mime_types() -> list[str]:
        return [i.data().decode() for i in ImageReader.supportedMimeTypes()]

    @staticmethod
    def get_image_formats_for_mime_type(typ: datatypes.ByteArrayType) -> list[str]:
        typ = datatypes.to_bytearray(typ)
        return [i.data().decode() for i in ImageReader.imageFormatsForMimeType(typ)]


if __name__ == "__main__":
    writer = ImageReader()
    writer.set_subtype("A8R8G8B8")



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from prettyqt import charts
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from collections.abc import Iterator


LabelsPositionStr = Literal["center", "on_value"]

LABELS_POSITIONS: bidict[LabelsPositionStr, charts.QCategoryAxis.AxisLabelsPosition] = (
    bidict(
        center=charts.QCategoryAxis.AxisLabelsPosition.AxisLabelsPositionCenter,
        on_value=charts.QCategoryAxis.AxisLabelsPosition.AxisLabelsPositionOnValue,
    )
)


class CategoryAxis(charts.ValueAxisMixin, charts.QCategoryAxis):
    def __delitem__(self, index: str):
        """Remove category label."""
        self.remove(index)

    def __getitem__(self, label: int | slice) -> str | list[str]:
        """Get category label from given index."""
        return self.categoriesLabels()[label]

    def __setitem__(self, index: str, value: str):
        """Set label at given index to value."""
        self.replaceLabel(index, value)

    def __iter__(self) -> Iterator[str]:
        """Iterate through all category labels."""
        return iter(self.categoriesLabels())

    def __add__(self, other: tuple[str, int]) -> CategoryAxis:
        """Append another category."""
        self.append(*other)
        return self

    def __len__(self):
        """Return category label count from axis."""
        return self.count()

    def set_labels_position(
        self, position: LabelsPositionStr | charts.QCategoryAxis.AxisLabelsPosition
    ):
        """Set the labels position.

        Args:
            position: labels position
        """
        self.setLabelsPosition(LABELS_POSITIONS.get_enum_value(position))

    def get_labels_position(self) -> LabelsPositionStr:
        """Return current labels position.

        Returns:
            labels position
        """
        return LABELS_POSITIONS.inverse[self.labelsPosition()]


if __name__ == "__main__":
    axis = CategoryAxis()



"""Tests for `prettyqt` package."""

# import pytest

import pytest


texttospeech = pytest.importorskip("prettyqt.texttospeech")


# from prettyqt.utils import InvalidParamError


def test_texttospeech(qtlog):
    with qtlog.disabled():
        tts = texttospeech.TextToSpeech()
    assert tts.get_state() in ["ready", "error"]
    tts.get_locale()
    tts.get_available_locales()
    tts.get_voice()
    tts.get_available_voices()


def test_voice():
    voice = texttospeech.Voice()
    assert voice.get_age() == "other"
    assert voice.get_gender() == "unknown"



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING

from prettyqt import core
from prettyqt.utils import datatypes, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterable


class FileSystemWatcher(core.ObjectMixin, core.QFileSystemWatcher):
    """Interface for monitoring files and directories for modifications."""

    def __repr__(self):
        return get_repr(self, self.directories() + self.files())

    def get_directories(self) -> list[pathlib.Path]:
        return [pathlib.Path(p) for p in self.directories()]

    def get_files(self) -> list[pathlib.Path]:
        return [pathlib.Path(p) for p in self.files()]

    def get_paths(self) -> list[pathlib.Path]:
        return self.get_directories() + self.get_files()

    def add_path(self, path: datatypes.PathType) -> bool:
        return self.addPath(os.fspath(path))

    def add_paths(self, paths: Iterable[datatypes.PathType]):
        self.addPaths([os.fspath(p) for p in paths])



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, datatypes


FloatingPointPrecisionStr = Literal["single", "double"]

FLOATING_POINT_PRECISION: bidict[
    FloatingPointPrecisionStr, QtCore.QDataStream.FloatingPointPrecision
] = bidict(
    single=QtCore.QDataStream.FloatingPointPrecision.SinglePrecision,
    double=QtCore.QDataStream.FloatingPointPrecision.DoublePrecision,
)

ByteOrderStr = Literal["big_endian", "little_endian"]

BYTE_ORDER: bidict[ByteOrderStr, QtCore.QDataStream.ByteOrder] = bidict(
    big_endian=QtCore.QDataStream.ByteOrder.BigEndian,
    little_endian=QtCore.QDataStream.ByteOrder.LittleEndian,
)

StatusStr = Literal["ok", "read_past_end", "read_corrupt_data", "write_failed"]

STATUS: bidict[StatusStr, QtCore.QDataStream.Status] = bidict(
    ok=QtCore.QDataStream.Status.Ok,
    read_past_end=QtCore.QDataStream.Status.ReadPastEnd,
    read_corrupt_data=QtCore.QDataStream.Status.ReadCorruptData,
    write_failed=QtCore.QDataStream.Status.WriteFailed,
)


class DataStream(QtCore.QDataStream):
    """Serialization of binary data to a QIODevice."""

    def set_byte_order(self, order: ByteOrderStr | QtCore.QDataStream.ByteOrder):
        """Set byte order.

        Args:
            order: byte order to use
        """
        self.setByteOrder(BYTE_ORDER.get_enum_value(order))

    def get_byte_order(self) -> ByteOrderStr:
        """Return byte order.

        Returns:
            byte order
        """
        return BYTE_ORDER.inverse[self.byteOrder()]

    def set_status(self, status: StatusStr):
        """Set status.

        Args:
            status: status to use
        """
        self.setStatus(STATUS.get_enum_value(status))

    def get_status(self) -> StatusStr:
        """Return status.

        Returns:
            status
        """
        return STATUS.inverse[self.status()]

    def set_floating_point_precision(
        self,
        precision: FloatingPointPrecisionStr | QtCore.QDataStream.FloatingPointPrecision,
    ):
        """Set floating point precision.

        Args:
            precision: floating point precision
        """
        self.setFloatingPointPrecision(FLOATING_POINT_PRECISION.get_enum_value(precision))

    def get_floating_point_precision(self) -> FloatingPointPrecisionStr:
        """Return floating point precision.

        Returns:
            floating point precision
        """
        return FLOATING_POINT_PRECISION.inverse[self.floatingPointPrecision()]

    @classmethod
    def create_bytearray(cls, data: datatypes.QtSerializableType) -> QtCore.QByteArray:
        ba = QtCore.QByteArray()
        stream = cls(ba, QtCore.QIODeviceBase.OpenModeFlag.WriteOnly)
        stream << data
        return ba

    @classmethod
    def write_bytearray(
        cls, ba: datatypes.ByteArrayType, write_to: datatypes.QtSerializableType
    ):
        ba = datatypes.to_bytearray(ba)
        stream = cls(ba, QtCore.QIODeviceBase.OpenModeFlag.ReadOnly)
        stream >> write_to

    @classmethod
    def copy_data(
        cls, source: datatypes.QtSerializableType, dest: datatypes.QtSerializableType
    ):
        ba = cls.create_bytearray(source)
        cls.write_bytearray(ba, dest)



from __future__ import annotations

from collections.abc import Iterator, Mapping, MutableMapping, Sequence
import contextlib
import logging
import os
import sys
from typing import Any, Literal, Self

from prettyqt import core
from prettyqt.utils import bidict, datatypes


logger = logging.getLogger(__name__)

FormatStr = Literal["user", "system"]

FORMAT: bidict[FormatStr, core.QSettings.Format] = bidict(
    native=core.QSettings.Format.NativeFormat, ini=core.QSettings.Format.IniFormat
)

ScopeStr = Literal["user", "system"]

SCOPE: bidict[ScopeStr, core.QSettings.Scope] = bidict(
    user=core.QSettings.Scope.UserScope, system=core.QSettings.Scope.SystemScope
)


class Settings_(  # noqa: N801
    core.ObjectMixin, core.QSettings, MutableMapping, metaclass=datatypes.QABCMeta
):
    """Persistent platform-independent application settings."""

    # Setting class with original behavior, compatible with original QSettings
    def __init__(self, *args, settings_id: str | None = None):
        super().__init__(*args)
        self.settings_id = settings_id

    def __repr__(self):
        return f"{type(self).__name__}({dict(self)})"

    def __contains__(self, key: str) -> bool:
        return self.contains(key)

    def __enter__(self):
        if self.settings_id:
            self.beginGroup(self.settings_id)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.settings_id:
            self.endGroup()

    def __getitem__(self, index: str):
        if index not in self.allKeys():
            raise KeyError(index)
        return self.get_value(index)

    def __setitem__(self, name: str, value):
        return self.set_value(name, value)

    def __delitem__(self, key: str):
        if not self.contains(key):
            raise KeyError(key)
        return self.remove(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.allKeys())

    def __len__(self) -> int:
        return len(self.allKeys())

    def __ior__(self, other: Mapping):
        self.update(other)
        return self

    def set_value(self, key: str, value):
        if not self.applicationName():
            msg = "no app name defined"
            raise RuntimeError(msg)
        if isinstance(value, Settings_):
            value = dict(value)
        self.setValue(key, value)

    def get_value(self, key: str, default=None):
        return self.value(key) if self.contains(key) else default

    @classmethod
    def set_default_format(cls, fmt: FormatStr | core.QSettings.Format):
        """Set the default format.

        Args:
            fmt: the default format to use
        """
        cls.setDefaultFormat(FORMAT.get_enum_value(fmt))

    @classmethod
    def get_default_format(cls) -> FormatStr:
        """Return default settings format.

        Returns:
            default settings format
        """
        return FORMAT.inverse[cls.defaultFormat()]

    def get_scope(self) -> ScopeStr:
        """Return scope.

        Returns:
            scope
        """
        return SCOPE.inverse[self.scope()]

    @classmethod
    def set_path(
        cls,
        fmt: FormatStr | core.QSettings.Format,
        scope: ScopeStr | core.QSettings.Scope,
        path: datatypes.PathType,
    ):
        """Set the path to the settings file.

        Args:
            fmt: the default format to use
            scope: the scope to use
            path: the path to set
        """
        fmt_val = FORMAT.get_enum_value(fmt)
        scope_val = SCOPE.get_enum_value(scope)
        cls.setPath(fmt_val, scope_val, os.fspath(path))

    @contextlib.contextmanager
    def edit_group(self, prefix: str):
        """Context manager for setting groups.

        Args:
            prefix: setting prefix for group
        """
        self.beginGroup(prefix)
        yield None
        self.endGroup()

    def write_array(self, prefix: str, array: Sequence) -> Iterator[tuple[Self, Any]]:
        """Context manager for writing arrays.

        Args:
            prefix: prefix for settings array
            array: array used to write settings

        Returns:
            Iterator yielding (setting, arrayitem) items
        """
        self.beginWriteArray(prefix, len(array))
        for i, _ in enumerate(array):
            self.setArrayIndex(i)
            yield (self, array[i])
        self.endArray()

    def read_array(self, prefix: str) -> Iterator[Settings_]:
        """Context manager for reading arrays.

        Args:
            prefix: prefix for settings array
        """
        size = self.beginReadArray(prefix)
        for i in range(size):
            self.setArrayIndex(i)
            yield self
        self.endArray()

    # MutableMapping overrides

    def keys(self) -> list[str]:
        return self.allKeys()

    def setdefault(self, key: str, default: Any = None) -> Any:
        if not self.contains(key):
            self.set_value(key, default)
            return default
        return self.get_value(key)

    @classmethod
    def register_extensions(
        cls,
        *exts: str,
        app_name: str | None = None,
        app_path: None | datatypes.PathType = None,
    ):
        logger.debug("assigning extensions %s to %r", exts, app_name)
        s = cls("HKEY_CURRENT_USER\\SOFTWARE\\Classes", Settings.Format.NativeFormat)
        if app_path is None:
            app_path = str(core.CoreApplication.get_application_file_path())
        app_path = os.fspath(app_path)
        if app_name is None:
            app_name = core.CoreApplication.applicationName()
        for ext in exts:
            s.setValue(f"{ext}/DefaultIcon/.", app_path)  # perhaps ,0 after app_path
            s.setValue(f"{ext}/.", app_name)
        if getattr(sys, "frozen", False):
            s.setValue(f"{app_name}/shell/open/command/.", f"{app_path} %1")
        else:
            command_value = rf'{sys.executable} -m {app_path} "%V"'
            s.setValue(f"{app_name}/shell/open/command/.", f"{command_value} %1")


class Settings(Settings_):
    """Settings class which wraps everything into a dict to preserve data types."""

    def set_value(self, key: str, value):
        match value:
            case Settings_():
                setting = dict(value=dict(value), typ="subsetting")
            case _:
                setting = dict(value=value, typ="regular")
        super().set_value(key, setting)

    def get_value(self, key: str, default=None):
        if not self.contains(key):
            return default
        val = self.value(key)
        # TODO: convert settings dicts back?
        match val:
            case {"value": setting}:
                return setting
            case _:
                # this is for migration
                self.set_value(key, val)
                return val


if __name__ == "__main__":
    settings = Settings("1", "2")
    settings["1"] = True
    settings |= dict(abc="abc")



from __future__ import annotations

from prettyqt import gui, widgets


class TracebackDialog(widgets.Dialog):
    """A dialog box that shows Python traceback."""

    def __init__(self, parent: widgets.QWidget | None = None):
        super().__init__(parent, window_title="Traceback")
        layout = self.set_layout("vertical")
        self._text = widgets.TextEdit(self, read_only=True, line_wrap_mode="none")
        self._text.setFontFamily(gui.Font.mono().family())
        layout.addWidget(self._text)
        self.resize(600, 400)

    def setText(self, text: str):
        """Always set text as a HTML text."""
        self._text.setHtml(text)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    wnd = TracebackDialog()
    wnd.show()
    with app.debug_mode():
        app.exec()



"""Tests for `prettyqt` package."""

import sys

import pytest

import prettyqt
from prettyqt import printsupport
from prettyqt.utils import InvalidParamError


pytestmark = pytest.mark.skipif(
    sys.platform == "linux" and prettyqt.qt.API.startswith("pyside"),
    reason="Segmentation fault",
)


def test_printer():
    printer = printsupport.Printer()
    assert printer.get_duplex() == "none"
    printer.set_pdf_version("v1_6")
    assert printer.get_pdf_version() == "v1_6"
    with pytest.raises(InvalidParamError):
        printer.set_pdf_version("test")


def test_printdialog():
    dlg = printsupport.PrintDialog()
    assert dlg is not None


def test_pagesetupdialog():
    dlg = printsupport.PageSetupDialog()
    assert dlg is not None


def test_printpreviewdialog():
    dlg = printsupport.PrintPreviewDialog()
    assert dlg is not None


def test_printpreviewwidget():
    widget = printsupport.PrintPreviewWidget()
    widget.set_view_mode("facing_pages")
    assert widget.get_view_mode() == "facing_pages"
    widget.set_zoom_mode("fit_to_width")
    assert widget.get_zoom_mode() == "fit_to_width"



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtPositioning
from prettyqt.utils import bidict


SatelliteSystemStr = Literal["undefined", "gps", "glonass"]

SATELLITE_SYSTEMS: bidict[
    SatelliteSystemStr, QtPositioning.QGeoSatelliteInfo.SatelliteSystem
] = bidict(
    undefined=QtPositioning.QGeoSatelliteInfo.SatelliteSystem.Undefined,
    gps=QtPositioning.QGeoSatelliteInfo.SatelliteSystem.GPS,
    glonass=QtPositioning.QGeoSatelliteInfo.SatelliteSystem.GLONASS,
)

AttributeStr = Literal["elevation", "azimuth"]

ATTRIBUTE: bidict[AttributeStr, QtPositioning.QGeoSatelliteInfo.Attribute] = bidict(
    elevation=QtPositioning.QGeoSatelliteInfo.Attribute.Elevation,
    azimuth=QtPositioning.QGeoSatelliteInfo.Attribute.Azimuth,
)


class GeoSatelliteInfo(QtPositioning.QGeoSatelliteInfo):
    def __getitem__(self, index: AttributeStr):
        return self.attribute(ATTRIBUTE[index])

    def __setitem__(self, index: AttributeStr, value: float):
        self.setAttribute(ATTRIBUTE[index], value)

    def __delitem__(self, index: AttributeStr):
        self.removeAttribute(ATTRIBUTE[index])

    def __contains__(self, value: AttributeStr):
        return self.hasAttribute(ATTRIBUTE[value])

    def __int__(self):
        return self.satelliteIdentifier()

    def set_satellite_system(
        self, system: SatelliteSystemStr | QtPositioning.QGeoSatelliteInfo.SatelliteSystem
    ):
        """Set satellite system.

        Args:
            system: satellite system to use
        """
        self.setSatelliteSystem(SATELLITE_SYSTEMS.get_enum_value(system))

    def get_satellite_system(self) -> SatelliteSystemStr:
        """Return satellite system.

        Returns:
            satellite system
        """
        return SATELLITE_SYSTEMS.inverse[self.satelliteSystem()]


if __name__ == "__main__":
    coord = GeoSatelliteInfo()
